# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io

from octoprint.util import atomic_write

class AbstractFileWrapper(object):
	"""
	Wrapper for file representations to save to storages.

	Arguments:
	    filename (str): The file's name
	"""

	def __init__(self, filename):
		self.filename = filename

	def save(self, path):
		"""
		Saves the file's content to the given absolute path.

		Arguments:
		    path (str): The absolute path to where to save the file
		"""
		raise NotImplementedError()

	def stream(self):
		"""
		Returns a Python stream object (subclass of io.IOBase) representing the file's contents.

		Returns:
		    io.IOBase: The file's contents as a stream.
		"""
		raise NotImplementedError()

class DiskFileWrapper(AbstractFileWrapper):
	"""
	An implementation of :class:`.AbstractFileWrapper` that wraps an actual file on disk. The `save` implementations
	will either copy the file to the new path (preserving file attributes) or -- if `move` is `True` (the default) --
	move the file.

	Arguments:
	    filename (str): The file's name
	    path (str): The file's absolute path
	    move (boolean): Whether to move the file upon saving (True, default) or copying.
	"""

	def __init__(self, filename, path, move=True):
		AbstractFileWrapper.__init__(self, filename)
		self.path = path
		self.move = move

	def save(self, path):
		import shutil

		if self.move:
			shutil.move(self.path, path)
		else:
			shutil.copy2(self.path, path)

	def stream(self):
		return io.open(self.path, 'rb')

class StreamWrapper(AbstractFileWrapper):
	"""
	A wrapper allowing processing of one or more consecutive streams.

	Arguments:
	    *streams (io.IOBase): One or more streams to process one after another to save to storage.
	"""
	def __init__(self, filename, *streams):
		if not len(streams) > 0:
			raise ValueError("Need at least one stream to wrap")

		AbstractFileWrapper.__init__(self, filename)
		self.streams = streams

	def save(self, path):
		"""
		Will dump the contents of all streams provided during construction into the target file, in the order they were
		provided.
		"""
		import shutil

		with atomic_write(path, mode='wb') as dest:
			with self.stream() as source:
				shutil.copyfileobj(source, dest)

	def stream(self):
		"""
		If more than one stream was provided to the constructor, will return a :class:`.MultiStream` wrapping all
		provided streams in the order they were provided, else the first and only stream is returned directly.
		"""
		if len(self.streams) > 1:
			return MultiStream(*self.streams)
		else:
			return self.streams[0]

class MultiStream(io.RawIOBase):
	"""
	A stream implementation which when read reads from multiple streams, one after the other, basically concatenating
	their contents in the order they are provided to the constructor.

	Arguments:
	    *streams (io.RawIOBase): One or more streams to concatenate.
	"""
	def __init__(self, *streams):
		io.RawIOBase.__init__(self)
		self.streams = streams
		self.current_stream = 0

	def read(self, n=-1):
		if n == 0:
			return b''

		if len(self.streams) == 0:
			return b''

		while self.current_stream < len(self.streams):
			stream = self.streams[self.current_stream]

			result = stream.read(n)
			if result is None or len(result) != 0:
				return result
			else:
				self.current_stream += 1

		return b''

	def readinto(self, b):
		n = len(b)
		read = self.read(n)
		b[:len(read)] = read
		return len(read)

	def close(self):
		for stream in self.streams:
			try:
				stream.close()
			except Exception:
				pass

	def readable(self, *args, **kwargs):
		return True

	def seekable(self, *args, **kwargs):
		return False

	def writable(self, *args, **kwargs):
		return False

class LineProcessorStream(io.RawIOBase):
	"""
	While reading from this stream the provided `input_stream` is read line by line, calling the (overridable) method
	:meth:`.process_line` for each read line.

	Sub classes can thus modify the contents of the `input_stream` in line, while it is being read. Keep in mind that
	``process_line`` will receive the line as a byte stream - if underlying code needs to operate on unicode you'll need
	to do the decoding yourself.

	Arguments:
	    input_stream (io.RawIOBase): The stream to process on the fly.
	"""

	def __init__(self, input_stream):
		io.RawIOBase.__init__(self)
		self.input_stream = io.BufferedReader(input_stream)
		self.leftover = None

	def read(self, n=-1):
		if n == 0:
			return b''

		result = b''
		while len(result) < n or n == -1:
			bytes_left = (n - len(result)) if n != -1 else -1
			if self.leftover is not None:
				if bytes_left != -1 and bytes_left < len(self.leftover):
					result += self.leftover[:bytes_left]
					self.leftover = self.leftover[bytes_left:]
					break
				else:
					result += self.leftover
					self.leftover = None

			processed_line = None
			while processed_line is None:
				line = self.input_stream.readline()
				if not line:
					break
				processed_line = self.process_line(line)

			if processed_line is None:
				break

			bytes_left = (n - len(result)) if n != -1 else -1
			if bytes_left != -1 and bytes_left < len(processed_line):
				result += processed_line[:bytes_left]
				self.leftover = processed_line[bytes_left:]
				break
			else:
				result += processed_line

		return result

	def readinto(self, b):
		n = len(b)
		read = self.read(n)
		b[:len(read)] = read
		return len(read)

	def process_line(self, line):
		"""
		Called from the `read` Method of this stream with each line read from `self.input_stream`.

		By returning ``None`` the line will not be returned from the read stream, effectively being stripped from the
		wrapper `input_stream`.

		Arguments:
		    line (bytes): The line as read from `self.input_stream` in byte representation (str under Python 2 and
		      bytes under Python 3)

		Returns:
		    bytes or None: The processed version of the line (might also be multiple lines), or None if the line is to be
		        stripped from the processed stream.
		"""
		return line

	def close(self):
		self.input_stream.close()

	def readable(self, *args, **kwargs):
		return True

	def seekable(self, *args, **kwargs):
		return False

	def writable(self, *args, **kwargs):
		return False



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
    sockjs.tornado.transports.websocket
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Websocket transport implementation
"""
import logging
import socket

from octoprint.vendor.sockjs.tornado import proto, websocket
from octoprint.vendor.sockjs.tornado.transports import base
from octoprint.vendor.sockjs.tornado.util import bytes_to_str
from tornado.websocket import WebSocketError
from tornado.ioloop import IOLoop

LOG = logging.getLogger("tornado.general")

class WebSocketTransport(websocket.SockJSWebSocketHandler, base.BaseTransportMixin):
    """Websocket transport"""
    name = 'websocket'

    def initialize(self, server):
        self.server = server
        self.session = None
        self.active = True

    def open(self, session_id):
        # Stats
        self.server.stats.on_conn_opened()

        # Disable nagle
        if self.server.settings['disable_nagle']:
            self.stream.set_nodelay(True)

        # Handle session
        self.session = self.server.create_session(session_id, register=False)

        if not self.session.set_handler(self):
            self.close()
            return

        self.session.verify_state()

        if self.session:
            self.session.flush()

    def _detach(self):
        if self.session is not None:
            self.session.remove_handler(self)
            self.session = None

    def on_message(self, message):
        # SockJS requires that empty messages should be ignored
        if not message or not self.session:
            return

        try:
            msg = proto.json_decode(bytes_to_str(message))

            if isinstance(msg, list):
                self.session.on_messages(msg)
            else:
                self.session.on_messages((msg,))
        except Exception:
            LOG.exception('WebSocket')

            # Close session on exception
            #self.session.close()

            # Close running connection
            self.abort_connection()

    def on_close(self):
        # Close session if websocket connection was closed
        if self.session is not None:
            # Stats
            self.server.stats.on_conn_closed()

            # Detach before closing session
            session = self.session
            self._detach()
            session.close()

    def send_pack(self, message, binary=False):
        if IOLoop.current(False) == self.server.io_loop:
            # Running in Main Thread
            # Send message
            try:
                self.write_message(message, binary)
            except (IOError, WebSocketError):
                self.server.io_loop.add_callback(self.on_close)
        else:
            # Not running in Main Thread so use proper thread to send message
            self.server.io_loop.add_callback(lambda: self.send_pack(message, binary))

    def session_closed(self):
        # If session was closed by the application, terminate websocket
        # connection as well.
        try:
            self.close()
        except IOError:
            pass
        finally:
            self._detach()

    # Websocket overrides
    def allow_draft76(self):
        return True

    def auto_decode(self):
        return False



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import socket
import sys
import netaddr
import netifaces
import logging

_cached_check_v6 = None
def check_v6():
	global _cached_check_v6

	def f():
		if not socket.has_ipv6:
			return False

		try:
			socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
		except Exception:
			# "[Errno 97] Address family not supported by protocol" or anything else really...
			return False
		return True

	if _cached_check_v6 is None:
		_cached_check_v6 = f()
	return _cached_check_v6

HAS_V6 = check_v6()

if hasattr(socket, "IPPROTO_IPV6") and hasattr(socket, "IPV6_V6ONLY"):
	# Dual stack support, hooray!
	IPPROTO_IPV6 = socket.IPPROTO_IPV6
	IPV6_V6ONLY = socket.IPV6_V6ONLY
else:
	if sys.platform == "win32":
		# Python 2.7 on Windows lacks IPPROTO_IPV6, but supports the socket options just fine, let's redefine it
		IPPROTO_IPV6 = 41
		IPV6_V6ONLY = 27
	else:
		# Whatever we are running on here, we don't want to use V6 on here due to lack of dual stack support
		HAS_V6 = False

def is_lan_address(address, additional_private=None):
	if not address:
		# no address is LAN address
		return True

	try:
		address = unmap_v4_as_v6(address)
		address = strip_interface_tag(address)

		ip = netaddr.IPAddress(address)
		if ip.is_private() or ip.is_loopback():
			return True

		if additional_private is None or not isinstance(additional_private, (list, tuple)):
			additional_private = []

		subnets = netaddr.IPSet()
		for additional in additional_private:
			try:
				subnets.add(netaddr.IPNetwork(additional))
			except Exception:
				if logging.getLogger(__name__).isEnabledFor(logging.DEBUG):
					logging.getLogger(__name__).exception("Error while trying to add additional private network to local subnets: {}".format(additional))

		def to_ipnetwork(address):
			prefix = address["netmask"]
			if "/" in prefix:
				# v6 notation in netifaces output, e.g. "ffff:ffff:ffff:ffff::/64"
				_, prefix = prefix.split("/")

			addr = strip_interface_tag(address["addr"])
			return netaddr.IPNetwork("{}/{}".format(addr, prefix))

		for interface in netifaces.interfaces():
			addrs = netifaces.ifaddresses(interface)
			for v4 in addrs.get(socket.AF_INET, ()):
				try:
					subnets.add(to_ipnetwork(v4))
				except Exception:
					if logging.getLogger(__name__).isEnabledFor(logging.DEBUG):
						logging.getLogger(__name__).exception("Error while trying to add v4 network to local subnets: {!r}".format(v4))

			if HAS_V6:
				for v6 in addrs.get(socket.AF_INET6, ()):
					try:
						subnets.add(to_ipnetwork(v6))
					except Exception:
						if logging.getLogger(__name__).isEnabledFor(logging.DEBUG):
							logging.getLogger(__name__).exception("Error while trying to add v6 network to local subnets: {!r}".format(v6))

		if ip in subnets:
			return True

		return False

	except Exception:
		# we are extra careful here since an unhandled exception in this method will effectively nuke the whole UI
		logging.getLogger(__name__).exception("Error while trying to determine whether {} is a local address".format(address))
		return True


def strip_interface_tag(address):
	if "%" in address:
		# interface comment, e.g. "fe80::457f:bbee:d579:1063%wlan0"
		address = address[:address.find("%")]
	return address


def unmap_v4_as_v6(address):
	if address.lower().startswith("::ffff:") and "." in address:
		# ipv6 mapped ipv4 address, unmap
		address = address[len("::ffff:"):]
	return address



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import io
from builtins import range

def readHex(filename):
	data = []
	extraAddr = 0
	f = io.open(filename, 'rt', encoding='utf-8')
	for line in f:
		line = line.strip()
		if line[0] != ':':
			raise Exception("Hex file has a line not starting with ':'")
		recLen = int(line[1:3], 16)
		addr = int(line[3:7], 16) + extraAddr
		recType = int(line[7:9], 16)
		if len(line) != recLen * 2 + 11:
			raise Exception("Error in hex file: " + line)
		checkSum = 0
		for i in range(0, recLen + 5):
			checkSum += int(line[i*2+1:i*2+3], 16)
		checkSum &= 0xFF
		if checkSum != 0:
			raise Exception("Checksum error in hex file: " + line)
		
		if recType == 0:#Data record
			while len(data) < addr + recLen:
				data.append(0)
			for i in range(0, recLen):
				data[addr + i] = int(line[i*2+9:i*2+11], 16)
		elif recType == 1:	#End Of File record
			pass
		elif recType == 2:	#Extended Segment Address Record
			extraAddr = int(line[9:13], 16) * 16
		else:
			print(recType, recLen, addr, checkSum, line)
	f.close()
	return data


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import os
import datetime
import io

from past.builtins import basestring

from collections import defaultdict
from flask import request, g, url_for, make_response, render_template, send_from_directory, redirect, abort, Response

import octoprint.plugin

from octoprint.server import app, userManager, groupManager, pluginManager, gettext, \
	debug, LOCALES, VERSION, DISPLAY_VERSION, BRANCH, preemptiveCache, \
	NOT_MODIFIED
from octoprint.access.permissions import Permissions
from octoprint.settings import settings
from octoprint.filemanager import full_extension_tree, get_all_extensions
from octoprint.util import to_unicode, to_bytes, sv

import re
import base64

from . import util

import logging
_logger = logging.getLogger(__name__)

_templates = dict()
_plugin_names = None
_plugin_vars = None

_valid_id_re = re.compile("[a-z_]+")
_valid_div_re = re.compile("[a-zA-Z_-]+")

def _preemptive_unless(base_url=None, additional_unless=None):
	if base_url is None:
		base_url = request.url_root

	disabled_for_root = not settings().getBoolean(["devel", "cache", "preemptive"]) \
	                    or base_url in settings().get(["server", "preemptiveCache", "exceptions"]) \
	                    or not (base_url.startswith("http://") or base_url.startswith("https://"))

	recording_disabled = request.headers.get("X-Preemptive-Record", "yes") == "no"

	if callable(additional_unless):
		return recording_disabled or disabled_for_root or additional_unless()
	else:
		return recording_disabled or disabled_for_root

def _preemptive_data(key, path=None, base_url=None, data=None, additional_request_data=None):
	if path is None:
		path = request.path
	if base_url is None:
		base_url = request.url_root

	d = dict(path=path,
	         base_url=base_url,
	         query_string="l10n={}".format(g.locale.language if g.locale else "en"))

	if key != "_default":
		d["plugin"] = key

	# add data if we have any
	if data is not None:
		try:
			if callable(data):
				data = data()
			if data:
				if "query_string" in data:
					data["query_string"] = "l10n={}&{}".format(g.locale.language, data["query_string"])
				d.update(data)
		except Exception:
			_logger.exception("Error collecting data for preemptive cache from plugin {}".format(key))

	# add additional request data if we have any
	if callable(additional_request_data):
		try:
			ard = additional_request_data()
			if ard:
				d.update(dict(
					_additional_request_data=ard
				))
		except Exception:
			_logger.exception("Error retrieving additional data for preemptive cache from plugin {}".format(key))

	return d

def _cache_key(ui, url=None, locale=None, additional_key_data=None):
	if url is None:
		url = request.base_url
	if locale is None:
		locale = g.locale.language if g.locale else "en"

	k = "ui:{}:{}:{}".format(ui, url, locale)
	if callable(additional_key_data):
		try:
			ak = additional_key_data()
			if ak:
				# we have some additional key components, let's attach them
				if not isinstance(ak, (list, tuple)):
					ak = [ak]
				k = "{}:{}".format(k, ":".join(ak))
		except Exception:
			_logger.exception("Error while trying to retrieve additional cache key parts for ui {}".format(ui))
	return k

def _valid_status_for_cache(status_code):
	return 200 <= status_code < 400

@app.route("/cached.gif")
def in_cache():
	url = request.base_url.replace("/cached.gif", "/")
	path = request.path.replace("/cached.gif", "/")
	base_url = request.url_root

	# select view from plugins and fall back on default view if no plugin will handle it
	ui_plugins = pluginManager.get_implementations(octoprint.plugin.UiPlugin,
	                                               sorting_context="UiPlugin.on_ui_render")
	for plugin in ui_plugins:
		try:
			if plugin.will_handle_ui(request):
				ui = plugin._identifier
				key = _cache_key(plugin._identifier,
				                 url=url,
				                 additional_key_data=plugin.get_ui_additional_key_data_for_cache)
				unless = _preemptive_unless(url, additional_unless=plugin.get_ui_preemptive_caching_additional_unless)
				data = _preemptive_data(plugin._identifier,
				                        path=path,
				                        base_url=base_url,
				                        data=plugin.get_ui_data_for_preemptive_caching,
				                        additional_request_data=plugin.get_ui_additional_request_data_for_preemptive_caching)
				break
		except Exception:
			_logger.exception("Error while calling plugin {}, skipping it".format(plugin._identifier),
			                  extra=dict(plugin=plugin._identifier))
	else:
		ui = "_default"
		key = _cache_key("_default", url=url)
		unless = _preemptive_unless(url)
		data = _preemptive_data("_default", path=path, base_url=base_url)

	response = make_response(bytes(base64.b64decode("R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7")))
	response.headers["Content-Type"] = "image/gif"

	if unless or not preemptiveCache.has_record(data, root=path):
		_logger.info("Preemptive cache not active for path {}, ui {} and data {!r}, signaling as cached".format(path, ui, data))
		return response
	elif util.flask.is_in_cache(key):
		_logger.info("Found path {} in cache (key: {}), signaling as cached".format(path, key))
		return response
	elif util.flask.is_cache_bypassed(key):
		_logger.info("Path {} was bypassed from cache (key: {}), signaling as cached".format(path, key))
		return response
	else:
		_logger.debug("Path {} not yet cached (key: {}), signaling as missing".format(path, key))
		return abort(404)

@app.route("/")
def index():
	from octoprint.server import printer

	global _templates, _plugin_names, _plugin_vars

	preemptive_cache_enabled = settings().getBoolean(["devel", "cache", "preemptive"])

	locale = g.locale.language if g.locale else "en"

	# helper to check if wizards are active
	def wizard_active(templates):
		return templates is not None and bool(templates["wizard"]["order"])

	# we force a refresh if the client forces one and we are not printing or if we have wizards cached
	client_refresh = util.flask.cache_check_headers()
	request_refresh = "_refresh" in request.values
	printing = printer.is_printing()
	if client_refresh and printing:
		logging.getLogger(__name__).warn("Client requested cache refresh via cache-control headers but we are printing. "
		                                 "Not invalidating caches due to resource limitation. Append ?_refresh=true to "
		                                 "the URL if you absolutely require a refresh now")
	client_refresh = client_refresh and not printing
	force_refresh = client_refresh \
	                or request_refresh \
	                or wizard_active(_templates.get(locale))

	# if we need to refresh our template cache or it's not yet set, process it
	fetch_template_data(refresh=force_refresh)

	now = datetime.datetime.utcnow()

	enable_accesscontrol = userManager.enabled
	enable_gcodeviewer = settings().getBoolean(["gcodeViewer", "enabled"])
	enable_timelapse = settings().getBoolean(["webcam", "timelapseEnabled"])
	enable_loading_animation = settings().getBoolean(["devel", "showLoadingAnimation"])

	def default_template_filter(template_type, template_key):
		if template_type == "navbar":
			return template_key != "login" or enable_accesscontrol
		elif template_type == "tab":
			return (template_key != "gcodeviewer" or enable_gcodeviewer) and \
			       (template_key != "timelapse" or enable_timelapse)
		elif template_type == "settings":
			return template_key != "accesscontrol" or enable_accesscontrol
		elif template_type == "usersettings":
			return enable_accesscontrol
		else:
			return True

	default_additional_etag = [enable_accesscontrol,
	                           enable_gcodeviewer,
	                           enable_timelapse,
	                           enable_loading_animation,
	                           wizard_active(_templates.get(locale))] + sorted(["{}:{}".format(to_unicode(k, errors="replace"),
	                                                                                           to_unicode(v, errors="replace"))
	                                                                           for k, v in _plugin_vars.items()])

	def get_preemptively_cached_view(key, view, data=None, additional_request_data=None, additional_unless=None):
		if (data is None and additional_request_data is None) or g.locale is None:
			return view

		d = _preemptive_data(key, data=data, additional_request_data=additional_request_data)

		def unless():
			return _preemptive_unless(base_url=request.url_root, additional_unless=additional_unless)

		# finally decorate our view
		return util.flask.preemptively_cached(cache=preemptiveCache,
		                                      data=d,
		                                      unless=unless)(view)

	def get_cached_view(key, view, additional_key_data=None, additional_files=None, additional_etag=None, custom_files=None, custom_etag=None, custom_lastmodified=None):
		if additional_etag is None:
			additional_etag = []

		def cache_key():
			return _cache_key(key, additional_key_data=additional_key_data)

		def check_etag_and_lastmodified():
			files = collect_files()
			lastmodified = compute_lastmodified(files)
			lastmodified_ok = util.flask.check_lastmodified(lastmodified)
			etag_ok = util.flask.check_etag(compute_etag(files=files,
			                                             lastmodified=lastmodified,
			                                             additional=[cache_key()] + additional_etag))
			return lastmodified_ok and etag_ok

		def validate_cache(cached):
			etag_different = compute_etag(additional=[cache_key()] + additional_etag) != cached.get_etag()[0]
			return force_refresh or etag_different

		def collect_files():
			if callable(custom_files):
				try:
					files = custom_files()
					if files:
						return files
				except Exception:
					_logger.exception("Error while trying to retrieve tracked files for plugin {}".format(key))

			templates = _get_all_templates()
			assets = _get_all_assets()
			translations = _get_all_translationfiles(g.locale.language if g.locale else "en",
			                                         "messages")

			files = templates + assets + translations

			if callable(additional_files):
				try:
					af = additional_files()
					if af:
						files += af
				except Exception:
					_logger.exception("Error while trying to retrieve additional tracked files for plugin {}".format(key))

			return sorted(set(files))

		def compute_lastmodified(files=None):
			if callable(custom_lastmodified):
				try:
					lastmodified = custom_lastmodified()
					if lastmodified:
						return lastmodified
				except Exception:
					_logger.exception("Error while trying to retrieve custom LastModified value for plugin {}".format(key))

			if files is None:
				files = collect_files()
			return _compute_date(files)

		def compute_etag(files=None, lastmodified=None, additional=None):
			if callable(custom_etag):
				try:
					etag = custom_etag()
					if etag:
						return etag
				except Exception:
					_logger.exception("Error while trying to retrieve custom ETag value for plugin {}".format(key))

			if files is None:
				files = collect_files()
			if lastmodified is None:
				lastmodified = compute_lastmodified(files)
			if lastmodified and not isinstance(lastmodified, basestring):
				from werkzeug.http import http_date
				lastmodified = http_date(lastmodified)
			if additional is None:
				additional = []

			import hashlib
			hash = hashlib.sha1()
			def hash_update(value):
				hash.update(to_bytes(value, encoding="utf-8", errors="replace"))
			hash_update(octoprint.__version__)
			hash_update(",".join(sorted(files)))
			if lastmodified:
				hash_update(lastmodified)
			for add in additional:
				hash_update(add)
			return hash.hexdigest()

		decorated_view = view
		decorated_view = util.flask.lastmodified(lambda _: compute_lastmodified())(decorated_view)
		decorated_view = util.flask.etagged(lambda _: compute_etag(additional=[cache_key()] + additional_etag))(decorated_view)
		decorated_view = util.flask.cached(timeout=-1,
		                                   refreshif=validate_cache,
		                                   key=cache_key,
		                                   unless_response=lambda response: util.flask.cache_check_response_headers(response) or util.flask.cache_check_status_code(response, _valid_status_for_cache))(decorated_view)
		decorated_view = util.flask.with_client_revalidation(decorated_view)
		decorated_view = util.flask.conditional(check_etag_and_lastmodified, NOT_MODIFIED)(decorated_view)
		return decorated_view

	def plugin_view(p):
		cached = get_cached_view(p._identifier,
		                         p.on_ui_render,
		                         additional_key_data=p.get_ui_additional_key_data_for_cache,
		                         additional_files=p.get_ui_additional_tracked_files,
		                         custom_files=p.get_ui_custom_tracked_files,
		                         custom_etag=p.get_ui_custom_etag,
		                         custom_lastmodified=p.get_ui_custom_lastmodified,
		                         additional_etag=p.get_ui_additional_etag(default_additional_etag))

		if preemptive_cache_enabled and p.get_ui_preemptive_caching_enabled():
			view = get_preemptively_cached_view(p._identifier,
			                                    cached,
			                                    p.get_ui_data_for_preemptive_caching,
			                                    p.get_ui_additional_request_data_for_preemptive_caching,
			                                    p.get_ui_preemptive_caching_additional_unless)
		else:
			view = cached

		template_filter = p.get_ui_custom_template_filter(default_template_filter)
		if template_filter is not None and callable(template_filter):
			filtered_templates = _filter_templates(_templates[locale], template_filter)
		else:
			filtered_templates = _templates[locale]

		render_kwargs = _get_render_kwargs(filtered_templates,
		                                   _plugin_names,
		                                   _plugin_vars,
		                                   now)

		return view(now, request, render_kwargs)

	def default_view():
		filtered_templates = _filter_templates(_templates[locale], default_template_filter)

		wizard = wizard_active(filtered_templates)
		accesscontrol_active = enable_accesscontrol and userManager.has_been_customized()

		render_kwargs = _get_render_kwargs(filtered_templates,
		                                   _plugin_names,
		                                   _plugin_vars,
		                                   now)
		render_kwargs.update(dict(
			enableWebcam=settings().getBoolean(["webcam", "webcamEnabled"]) and bool(settings().get(["webcam", "stream"])),
			enableTemperatureGraph=settings().get(["feature", "temperatureGraph"]),
			enableAccessControl=enable_accesscontrol,
			accessControlActive=accesscontrol_active,
			enableLoadingAnimation=enable_loading_animation,
			enableSdSupport=settings().get(["feature", "sdSupport"]),
			gcodeMobileThreshold=settings().get(["gcodeViewer", "mobileSizeThreshold"]),
			gcodeThreshold=settings().get(["gcodeViewer", "sizeThreshold"]),
			wizard=wizard,
			now=now,
		))

		# no plugin took an interest, we'll use the default UI
		def make_default_ui():
			r = make_response(render_template("index.jinja2", **render_kwargs))
			if wizard:
				# if we have active wizard dialogs, set non caching headers
				r = util.flask.add_non_caching_response_headers(r)
			return r

		cached = get_cached_view("_default",
		                         make_default_ui,
		                         additional_etag=default_additional_etag)
		preemptively_cached = get_preemptively_cached_view("_default",
		                                                   cached,
		                                                   dict(),
		                                                   dict())
		return preemptively_cached()

	response = None

	forced_view = request.headers.get("X-Force-View", None)

	if forced_view:
		# we have view forced by the preemptive cache
		_logger.debug("Forcing rendering of view {}".format(forced_view))
		if forced_view != "_default":
			plugin = pluginManager.get_plugin_info(forced_view, require_enabled=True)
			if plugin is not None and isinstance(plugin.implementation, octoprint.plugin.UiPlugin):
				response = plugin_view(plugin.implementation)
				if _logger.isEnabledFor(logging.DEBUG) and isinstance(response, Response):
					response.headers["X-Ui-Plugin"] = plugin._identifier
		else:
			response = default_view()
			if _logger.isEnabledFor(logging.DEBUG) and isinstance(response, Response):
				response.headers["X-Ui-Plugin"] = "_default"

	else:
		# select view from plugins and fall back on default view if no plugin will handle it
		ui_plugins = pluginManager.get_implementations(octoprint.plugin.UiPlugin, sorting_context="UiPlugin.on_ui_render")
		for plugin in ui_plugins:
			try:
				if plugin.will_handle_ui(request):
					# plugin claims responsibility, let it render the UI
					response = plugin_view(plugin)
					if response is not None:
						if _logger.isEnabledFor(logging.DEBUG) and isinstance(response, Response):
							response.headers["X-Ui-Plugin"] = plugin._identifier
						break
					else:
						_logger.warning("UiPlugin {} returned an empty response".format(plugin._identifier))
			except Exception:
				_logger.exception("Error while calling plugin {}, skipping it".format(plugin._identifier),
				                  extra=dict(plugin=plugin._identifier))
		else:
			response = default_view()
			if _logger.isEnabledFor(logging.DEBUG) and isinstance(response, Response):
				response.headers["X-Ui-Plugin"] = "_default"

	if response is None:
		return abort(404)
	return response


def _get_render_kwargs(templates, plugin_names, plugin_vars, now):
	global _logger

	#~~ a bunch of settings

	first_run = settings().getBoolean(["server", "firstRun"])

	locales = dict()
	for l in LOCALES:
		try:
			locales[l.language] = dict(language=l.language, display=l.display_name, english=l.english_name)
		except Exception:
			_logger.exception("Error while collecting available locales")

	permissions = [permission.as_dict() for permission in Permissions.all()]
	filetypes = list(sorted(full_extension_tree().keys()))
	extensions = list(map(lambda ext: ".{}".format(ext), get_all_extensions()))

	#~~ prepare full set of template vars for rendering

	render_kwargs = dict(
		debug=debug,
		firstRun=first_run,
		version=dict(number=VERSION, display=DISPLAY_VERSION, branch=BRANCH),
		templates=templates,
		pluginNames=plugin_names,
		locales=locales,
		permissions=permissions,
		supportedFiletypes=filetypes,
		supportedExtensions=extensions
	)
	render_kwargs.update(plugin_vars)

	return render_kwargs


def fetch_template_data(refresh=False):
	global _templates, _plugin_names, _plugin_vars

	locale = g.locale.language if g.locale else "en"

	if not refresh and _templates.get(locale) is not None and _plugin_names is not None and _plugin_vars is not None:
		return _templates[locale], _plugin_names, _plugin_vars

	first_run = settings().getBoolean(["server", "firstRun"])

	##~~ prepare templates

	templates = defaultdict(lambda: dict(order=[], entries=dict()))

	# rules for transforming template configs to template entries
	template_rules = dict(
		navbar=dict(div=lambda x: "navbar_plugin_" + x, template=lambda x: x + "_navbar.jinja2", to_entry=lambda data: data),
		sidebar=dict(div=lambda x: "sidebar_plugin_" + x, template=lambda x: x + "_sidebar.jinja2", to_entry=lambda data: (data["name"], data)),
		tab=dict(div=lambda x: "tab_plugin_" + x, template=lambda x: x + "_tab.jinja2", to_entry=lambda data: (data["name"], data)),
		settings=dict(div=lambda x: "settings_plugin_" + x, template=lambda x: x + "_settings.jinja2", to_entry=lambda data: (data["name"], data)),
		usersettings=dict(div=lambda x: "usersettings_plugin_" + x, template=lambda x: x + "_usersettings.jinja2", to_entry=lambda data: (data["name"], data)),
		wizard=dict(div=lambda x: "wizard_plugin_" + x, template=lambda x: x + "_wizard.jinja2", to_entry=lambda data: (data["name"], data)),
		about=dict(div=lambda x: "about_plugin_" + x, template=lambda x: x + "_about.jinja2", to_entry=lambda data: (data["name"], data)),
		generic=dict(template=lambda x: x + ".jinja2", to_entry=lambda data: data)
	)

	# sorting orders
	def wizard_key_extractor(d, k):
		if d[1].get("_key", None) == "plugin_corewizard_acl":
			# Ultra special case - we MUST always have the ACL wizard first since otherwise any steps that follow and
			# that require to access APIs to function will run into errors since those APIs won't work before ACL
			# has been configured. See also #2140
			return "0:{}".format(to_unicode(d[0]))
		elif d[1].get("mandatory", False):
			# Other mandatory steps come before the optional ones
			return "1:{}".format(to_unicode(d[0]))
		else:
			# Finally everything else
			return "2:{}".format(to_unicode(d[0]))

	template_sorting = dict(
		navbar=dict(add="prepend", key=None),
		sidebar=dict(add="append", key="name"),
		tab=dict(add="append", key="name"),
		settings=dict(add="custom_append", key="name", custom_add_entries=lambda missing: dict(section_plugins=(gettext("Plugins"), None)), custom_add_order=lambda missing: ["section_plugins"] + missing),
		usersettings=dict(add="append", key="name"),
		wizard=dict(add="append", key="name", key_extractor=wizard_key_extractor),
		about=dict(add="append", key="name"),
		generic=dict(add="append", key=None)
	)

	hooks = pluginManager.get_hooks("octoprint.ui.web.templatetypes")
	for name, hook in hooks.items():
		try:
			result = hook(dict(template_sorting), dict(template_rules))
		except Exception:
			_logger.exception("Error while retrieving custom template type "
			                  "definitions from plugin {name}".format(**locals()),
			                  extra=dict(plugin=name))
		else:
			if not isinstance(result, list):
				continue

			for entry in result:
				if not isinstance(entry, tuple) or not len(entry) == 3:
					continue

				key, order, rule = entry

				# order defaults
				if "add" not in order:
					order["add"] = "prepend"
				if "key" not in order:
					order["key"] = "name"

				# rule defaults
				if "div" not in rule:
					# default div name: <hook plugin>_<template_key>_plugin_<plugin>
					div = "{name}_{key}_plugin_".format(**locals())
					rule["div"] = lambda x: div + x
				if "template" not in rule:
					# default template name: <plugin>_plugin_<hook plugin>_<template key>.jinja2
					template = "_plugin_{name}_{key}.jinja2".format(**locals())
					rule["template"] = lambda x: x + template
				if "to_entry" not in rule:
					# default to_entry assumes existing "name" property to be used as label for 2-tuple entry data structure (<name>, <properties>)
					rule["to_entry"] = lambda data: (data["name"], data)

				template_rules["plugin_" + name + "_" + key] = rule
				template_sorting["plugin_" + name + "_" + key] = order
	template_types = list(template_rules.keys())

	# navbar

	templates["navbar"]["entries"] = dict(
		settings=dict(template="navbar/settings.jinja2", _div="navbar_settings", styles=["display: none;"], data_bind="visible: loginState.hasPermissionKo(access.permissions.SETTINGS)"),
		systemmenu=dict(template="navbar/systemmenu.jinja2", _div="navbar_systemmenu", classes=["dropdown"], styles=["display: none;"], data_bind="visible: loginState.hasPermissionKo(access.permissions.SYSTEM)", custom_bindings=False),
		login=dict(template="navbar/login.jinja2", _div="navbar_login", classes=["dropdown"], custom_bindings=False),
	)

	# sidebar

	templates["sidebar"]["entries"]= dict(
		connection=(gettext("Connection"), dict(template="sidebar/connection.jinja2", _div="connection", icon="signal", styles_wrapper=["display: none;"], data_bind="visible: loginState.hasPermissionKo(access.permissions.CONNECTION)", template_header="sidebar/connection_header.jinja2")),
		state=(gettext("State"), dict(template="sidebar/state.jinja2", _div="state", icon="info-circle", styles_wrapper=["display: none;"], data_bind="visible: loginState.hasPermissionKo(access.permissions.STATUS)")),
		files=(gettext("Files"), dict(template="sidebar/files.jinja2", _div="files", icon="list", classes_content=["overflow_visible"], template_header="sidebar/files_header.jinja2", styles_wrapper=["display: none;"], data_bind="visible: loginState.hasPermissionKo(access.permissions.FILES_LIST)"))
	)

	# tabs

	templates["tab"]["entries"] = dict(
		temperature=(gettext("Temperature"), dict(template="tabs/temperature.jinja2", _div="temp", styles=["display: none;"], data_bind="visible: loginState.hasAnyPermissionKo(access.permissions.STATUS, access.permissions.CONTROL)() && visible()")),
		control=(gettext("Control"), dict(template="tabs/control.jinja2", _div="control", styles=["display: none;"], data_bind="visible: loginState.hasAnyPermissionKo(access.permissions.WEBCAM, access.permissions.CONTROL)")),
		gcodeviewer=(gettext("GCode Viewer"), dict(template="tabs/gcodeviewer.jinja2", _div="gcode", styles=["display: none;"], data_bind="visible: loginState.hasAllPermissionsKo(access.permissions.GCODE_VIEWER, access.permissions.FILES_DOWNLOAD)")),
		terminal=(gettext("Terminal"), dict(template="tabs/terminal.jinja2", _div="term", styles=["display: none;"], data_bind="visible: loginState.hasPermissionKo(access.permissions.MONITOR_TERMINAL)")),
		timelapse=(gettext("Timelapse"), dict(template="tabs/timelapse.jinja2", _div="timelapse", styles=["display: none;"], data_bind="visible: loginState.hasPermissionKo(access.permissions.TIMELAPSE_LIST)"))
	)

	# settings dialog

	templates["settings"]["entries"] = dict(
		section_printer=(gettext("Printer"), None),

		serial=(gettext("Serial Connection"), dict(template="dialogs/settings/serialconnection.jinja2", _div="settings_serialConnection", custom_bindings=False)),
		printerprofiles=(gettext("Printer Profiles"), dict(template="dialogs/settings/printerprofiles.jinja2", _div="settings_printerProfiles", custom_bindings=False)),
		temperatures=(gettext("Temperatures"), dict(template="dialogs/settings/temperatures.jinja2", _div="settings_temperature", custom_bindings=False)),
		terminalfilters=(gettext("Terminal Filters"), dict(template="dialogs/settings/terminalfilters.jinja2", _div="settings_terminalFilters", custom_bindings=False)),
		gcodescripts=(gettext("GCODE Scripts"), dict(template="dialogs/settings/gcodescripts.jinja2", _div="settings_gcodeScripts", custom_bindings=False)),

		section_features=(gettext("Features"), None),

		features=(gettext("Features"), dict(template="dialogs/settings/features.jinja2", _div="settings_features", custom_bindings=False)),
		webcam=(gettext("Webcam & Timelapse"), dict(template="dialogs/settings/webcam.jinja2", _div="settings_webcam", custom_bindings=False)),
		gcodevisualizer=(gettext("GCODE Visualizer"), dict(template="dialogs/settings/gcodevisualizer.jinja2", _div="settings_gcodegcodevisualizer", custom_bindings=False)),
		api=(gettext("API"), dict(template="dialogs/settings/api.jinja2", _div="settings_api", custom_bindings=False)),

		section_octoprint=(gettext("OctoPrint"), None),

		accesscontrol=(gettext("Access Control"), dict(template="dialogs/settings/accesscontrol.jinja2", _div="settings_users", custom_bindings=False)),
		folders=(gettext("Folders"), dict(template="dialogs/settings/folders.jinja2", _div="settings_folders", custom_bindings=False)),
		appearance=(gettext("Appearance"), dict(template="dialogs/settings/appearance.jinja2", _div="settings_appearance", custom_bindings=False)),
		server=(gettext("Server"), dict(template="dialogs/settings/server.jinja2", _div="settings_server", custom_bindings=False)),
	)

	# user settings dialog

	templates["usersettings"]["entries"] = dict(
		access=(gettext("Access"), dict(template="dialogs/usersettings/access.jinja2", _div="usersettings_access", custom_bindings=False)),
		interface=(gettext("Interface"), dict(template="dialogs/usersettings/interface.jinja2", _div="usersettings_interface", custom_bindings=False)),
	)

	# wizard

	if first_run:
		def custom_insert_order(existing, missing):
			if "firstrunstart" in missing:
				missing.remove("firstrunstart")
			if "firstrunend" in missing:
				missing.remove("firstrunend")

			return ["firstrunstart"] + existing + missing + ["firstrunend"]

		template_sorting["wizard"].update(dict(add="custom_insert", custom_insert_entries=lambda missing: dict(), custom_insert_order=custom_insert_order))
		templates["wizard"]["entries"] = dict(
			firstrunstart=(gettext("Start"), dict(template="dialogs/wizard/firstrun_start.jinja2", _div="wizard_firstrun_start")),
			firstrunend=(gettext("Finish"), dict(template="dialogs/wizard/firstrun_end.jinja2", _div="wizard_firstrun_end")),
		)

	# about dialog

	templates["about"]["entries"] = dict(
		about=("About OctoPrint", dict(template="dialogs/about/about.jinja2", _div="about_about", custom_bindings=False)),
		license=("OctoPrint License", dict(template="dialogs/about/license.jinja2", _div="about_license", custom_bindings=False)),
		thirdparty=("Third Party Licenses", dict(template="dialogs/about/thirdparty.jinja2", _div="about_thirdparty", custom_bindings=False)),
		authors=("Authors", dict(template="dialogs/about/authors.jinja2", _div="about_authors", custom_bindings=False)),
		supporters=("Supporters", dict(template="dialogs/about/supporters.jinja2", _div="about_sponsors", custom_bindings=False))
	)

	# extract data from template plugins

	template_plugins = pluginManager.get_implementations(octoprint.plugin.TemplatePlugin)

	plugin_vars = dict()
	plugin_names = set()
	seen_wizards = settings().get(["server", "seenWizards"]) if not first_run else dict()
	for implementation in template_plugins:
		name = implementation._identifier
		plugin_names.add(name)
		wizard_required = False
		wizard_ignored = False

		try:
			vars = implementation.get_template_vars()
			configs = implementation.get_template_configs()
			if isinstance(implementation, octoprint.plugin.WizardPlugin):
				wizard_required = implementation.is_wizard_required()
				wizard_ignored = octoprint.plugin.WizardPlugin.is_wizard_ignored(seen_wizards, implementation)
		except Exception:
			_logger.exception("Error while retrieving template data for plugin {}, ignoring it".format(name),
			                  extra=dict(plugin=name))
			continue

		if not isinstance(vars, dict):
			vars = dict()
		if not isinstance(configs, (list, tuple)):
			configs = []

		for var_name, var_value in vars.items():
			plugin_vars["plugin_" + name + "_" + var_name] = var_value

		includes = _process_template_configs(name, implementation, configs, template_rules)

		if not wizard_required or wizard_ignored:
			includes["wizard"] = list()

		for t in template_types:
			for include in includes[t]:
				if t == "navbar" or t == "generic":
					data = include
				else:
					data = include[1]

				key = data["_key"]
				if "replaces" in data:
					key = data["replaces"]
				templates[t]["entries"][key] = include

	#~~ order internal templates and plugins

	# make sure that
	# 1) we only have keys in our ordered list that we have entries for and
	# 2) we have all entries located somewhere within the order

	for t in template_types:
		default_order = settings().get(["appearance", "components", "order", t], merged=True, config=dict()) or []
		configured_order = settings().get(["appearance", "components", "order", t], merged=True) or []
		configured_disabled = settings().get(["appearance", "components", "disabled", t]) or []

		# first create the ordered list of all component ids according to the configured order
		templates[t]["order"] = [x for x in configured_order if x in templates[t]["entries"] and not x in configured_disabled]

		# now append the entries from the default order that are not already in there
		templates[t]["order"] += [x for x in default_order if not x in templates[t]["order"] and x in templates[t]["entries"] and not x in configured_disabled]

		all_ordered = set(templates[t]["order"])
		all_disabled = set(configured_disabled)

		# check if anything is missing, if not we are done here
		missing_in_order = set(templates[t]["entries"].keys()).difference(all_ordered).difference(all_disabled)
		if len(missing_in_order) == 0:
			continue

		# works with entries that are dicts and entries that are 2-tuples with the
		# entry data at index 1
		def config_extractor(item, key, default_value=None):
			if isinstance(item, dict) and key in item:
				return item[key] if key in item else default_value
			elif isinstance(item, tuple) and len(item) > 1 and isinstance(item[1], dict) and key in item[1]:
				return item[1][key] if key in item[1] else default_value

			return default_value

		# finally add anything that's not included in our order yet
		if template_sorting[t]["key"] is not None:
			# we'll use our config extractor as default key extractor
			extractor = config_extractor

			# if template type provides custom extractor, make sure its exceptions are handled
			if "key_extractor" in template_sorting[t] and callable(template_sorting[t]["key_extractor"]):
				def create_safe_extractor(extractor):
					def f(x, k):
						try:
							return extractor(x, k)
						except Exception:
							_logger.exception("Error while extracting sorting keys for template {}".format(t))
							return None
					return f
				extractor = create_safe_extractor(template_sorting[t]["key_extractor"])

			sort_key = template_sorting[t]["key"]

			def key_func(x):
				config = templates[t]["entries"][x]
				entry_order = config_extractor(config, "order", default_value=None)
				return entry_order is None, sv(entry_order), sv(extractor(config, sort_key))

			sorted_missing = sorted(missing_in_order, key=key_func)
		else:
			def key_func(x):
				config = templates[t]["entries"][x]
				entry_order = config_extractor(config, "order", default_value=None)
				return entry_order is None, sv(entry_order)

			sorted_missing = sorted(missing_in_order, key=key_func)

		if template_sorting[t]["add"] == "prepend":
			templates[t]["order"] = sorted_missing + templates[t]["order"]
		elif template_sorting[t]["add"] == "append":
			templates[t]["order"] += sorted_missing
		elif template_sorting[t]["add"] == "custom_prepend" and "custom_add_entries" in template_sorting[t] and "custom_add_order" in template_sorting[t]:
			templates[t]["entries"].update(template_sorting[t]["custom_add_entries"](sorted_missing))
			templates[t]["order"] = template_sorting[t]["custom_add_order"](sorted_missing) + templates[t]["order"]
		elif template_sorting[t]["add"] == "custom_append" and "custom_add_entries" in template_sorting[t] and "custom_add_order" in template_sorting[t]:
			templates[t]["entries"].update(template_sorting[t]["custom_add_entries"](sorted_missing))
			templates[t]["order"] += template_sorting[t]["custom_add_order"](sorted_missing)
		elif template_sorting[t]["add"] == "custom_insert" and "custom_insert_entries" in template_sorting[t] and "custom_insert_order" in template_sorting[t]:
			templates[t]["entries"].update(template_sorting[t]["custom_insert_entries"](sorted_missing))
			templates[t]["order"] = template_sorting[t]["custom_insert_order"](templates[t]["order"], sorted_missing)

	_templates[locale] = templates
	_plugin_names = plugin_names
	_plugin_vars = plugin_vars

	return templates, plugin_names, plugin_vars


def _process_template_configs(name, implementation, configs, rules):
	from jinja2.exceptions import TemplateNotFound

	counters = defaultdict(lambda: 1)
	includes = defaultdict(list)

	for config in configs:
		if not isinstance(config, dict):
			continue
		if not "type" in config:
			continue

		template_type = config["type"]
		del config["type"]

		if not template_type in rules:
			continue
		rule = rules[template_type]

		data = _process_template_config(name, implementation, rule, config=config, counter=counters[template_type])
		if data is None:
			continue

		includes[template_type].append(rule["to_entry"](data))
		counters[template_type] += 1

	for template_type in rules:
		if len(includes[template_type]) == 0:
			# if no template of that type was added by the config, we'll try to use the default template name
			rule = rules[template_type]
			data = _process_template_config(name, implementation, rule)
			if data is not None:
				try:
					app.jinja_env.get_or_select_template(data["template"])
				except TemplateNotFound:
					pass
				except Exception:
					_logger.exception("Error in template {}, not going to include it".format(data["template"]))
				else:
					includes[template_type].append(rule["to_entry"](data))

	return includes


def _process_template_config(name, implementation, rule, config=None, counter=1):
	if "mandatory" in rule:
		for mandatory in rule["mandatory"]:
			if not mandatory in config:
				return None

	if config is None:
		config = dict()
	data = dict(config)

	if not "suffix" in data and counter > 1:
		data["suffix"] = "_%d" % counter

	if "div" in data:
		data["_div"] = data["div"]
	elif "div" in rule:
		data["_div"] = rule["div"](name)
		if "suffix" in data:
			data["_div"] = data["_div"] + data["suffix"]
		if not _valid_div_re.match(data["_div"]):
			_logger.warning("Template config {} contains invalid div identifier {}, skipping it".format(name, data["_div"]))
			return None

	if not "template" in data:
		data["template"] = rule["template"](name)

	if not "name" in data:
		data["name"] = implementation._plugin_name

	if not "custom_bindings" in data or data["custom_bindings"]:
		data_bind = "allowBindings: true"
		if "data_bind" in data:
			data_bind = data_bind + ", " + data["data_bind"]
		data_bind = data_bind.replace("\"", "\\\"")
		data["data_bind"] = data_bind

	data["_key"] = "plugin_" + name
	if "suffix" in data:
		data["_key"] += data["suffix"]

	data["_plugin"] = name

	return data


def _filter_templates(templates, template_filter):
	filtered_templates = dict()
	for template_type, template_collection in templates.items():
		filtered_entries = dict()
		for template_key, template_entry in template_collection["entries"].items():
			if template_filter(template_type, template_key):
				filtered_entries[template_key] = template_entry
		filtered_templates[template_type] = dict(order=list(filter(lambda x: x in filtered_entries,
		                                                      	   template_collection["order"])),
		                                         entries=filtered_entries)
	return filtered_templates


@app.route("/robots.txt")
@util.flask.cached(timeout=-1)
def robotsTxt():
	return send_from_directory(app.static_folder, "robots.txt")


@app.route("/i18n/<string:locale>/<string:domain>.js")
@util.flask.conditional(lambda: _check_etag_and_lastmodified_for_i18n(), NOT_MODIFIED)
@util.flask.etagged(lambda _: _compute_etag_for_i18n(request.view_args["locale"], request.view_args["domain"]))
@util.flask.lastmodified(lambda _: _compute_date_for_i18n(request.view_args["locale"], request.view_args["domain"]))
def localeJs(locale, domain):
	messages = dict()
	plural_expr = None

	if locale != "en":
		messages, plural_expr = _get_translations(locale, domain)

	catalog = dict(
		messages=messages,
		plural_expr=plural_expr,
		locale=locale,
		domain=domain
	)

	from flask import Response
	return Response(render_template("i18n.js.jinja2", catalog=catalog), content_type="application/x-javascript; charset=utf-8")


@app.route("/plugin_assets/<string:name>/<path:filename>")
def plugin_assets(name, filename):
	return redirect(url_for("plugin." + name + ".static", filename=filename))


def _compute_etag_for_i18n(locale, domain, files=None, lastmodified=None):
	if files is None:
		files = _get_all_translationfiles(locale, domain)
	if lastmodified is None:
		lastmodified = _compute_date(files)
	if lastmodified and not isinstance(lastmodified, basestring):
		from werkzeug.http import http_date
		lastmodified = http_date(lastmodified)

	import hashlib
	hash = hashlib.sha1()
	def hash_update(value):
		hash.update(value.encode('utf-8'))
	hash_update(",".join(sorted(files)))
	if lastmodified:
		hash_update(lastmodified)
	return hash.hexdigest()


def _compute_date_for_i18n(locale, domain):
	return _compute_date(_get_all_translationfiles(locale, domain))


def _compute_date(files):
	from datetime import datetime
	timestamps = list(map(lambda path: os.stat(path).st_mtime, files)) + [0] if files else []
	max_timestamp = max(*timestamps) if timestamps else None
	if max_timestamp:
		# we set the micros to 0 since microseconds are not speced for HTTP
		max_timestamp = datetime.fromtimestamp(max_timestamp).replace(microsecond=0)
	return max_timestamp


def _check_etag_and_lastmodified_for_i18n():
	locale = request.view_args["locale"]
	domain = request.view_args["domain"]

	etag_ok = util.flask.check_etag(_compute_etag_for_i18n(request.view_args["locale"], request.view_args["domain"]))

	lastmodified = _compute_date_for_i18n(locale, domain)
	lastmodified_ok = lastmodified is None or util.flask.check_lastmodified(lastmodified)

	return etag_ok and lastmodified_ok


def _get_all_templates():
	from octoprint.util.jinja import get_all_template_paths
	return get_all_template_paths(app.jinja_loader)


def _get_all_assets():
	from octoprint.util.jinja import get_all_asset_paths
	return get_all_asset_paths(app.jinja_env.assets_environment)


def _get_all_translationfiles(locale, domain):
	from flask import _request_ctx_stack

	def get_po_path(basedir, locale, domain):
		path = os.path.join(basedir, locale)
		if not os.path.isdir(path):
			return None

		path = os.path.join(path, "LC_MESSAGES", "{domain}.po".format(**locals()))
		if not os.path.isfile(path):
			return None

		return path

	po_files = []

	user_base_path = os.path.join(settings().getBaseFolder("translations", check_writable=False))
	user_plugin_path = os.path.join(user_base_path, "_plugins")

	# plugin translations
	plugins = octoprint.plugin.plugin_manager().enabled_plugins
	for name, plugin in plugins.items():
		dirs = [os.path.join(user_plugin_path, name), os.path.join(plugin.location, 'translations')]
		for dirname in dirs:
			if not os.path.isdir(dirname):
				continue

			po_file = get_po_path(dirname, locale, domain)
			if po_file:
				po_files.append(po_file)
				break

	# core translations
	ctx = _request_ctx_stack.top
	base_path = os.path.join(ctx.app.root_path, "translations")

	dirs = [user_base_path, base_path]
	for dirname in dirs:
		po_file = get_po_path(dirname, locale, domain)
		if po_file:
			po_files.append(po_file)
			break

	return po_files


def _get_translations(locale, domain):
	from babel.messages.pofile import read_po
	from octoprint.util import dict_merge

	messages = dict()
	plural_expr = None

	def messages_from_po(path, locale, domain):
		messages = dict()
		with io.open(path, mode="rt", encoding="utf-8") as f:
			catalog = read_po(f, locale=locale, domain=domain)

			for message in catalog:
				message_id = message.id
				if isinstance(message_id, (list, tuple)):
					message_id = message_id[0]
				if message.string:
					messages[message_id] = message.string

		return messages, catalog.plural_expr

	po_files = _get_all_translationfiles(locale, domain)
	for po_file in po_files:
		po_messages, plural_expr = messages_from_po(po_file, locale, domain)
		if po_messages is not None:
			messages = dict_merge(messages, po_messages)

	return messages, plural_expr



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from octoprint.server.api import api
from flask import redirect, url_for

# NOTE: The redirects here should rather be 308 PERMANENT REDIRECT, however RFC7538 doesn't seem to be supported
# by all browsers yet. So we stick to 307 TEMPORARY REDIRECT as defined in RFC7231 although it's definitely not a
# temporary redirect we have here.

@api.route("/users", methods=["GET"])
def deprecated_get_users():
	return redirect(url_for("api.get_users"), code=307)


@api.route("/users", methods=["POST"])
def addUser():
	return redirect(url_for("api.add_user"), code=307)


@api.route("/users/<username>", methods=["GET"])
def getUser(username):
	return redirect(url_for("api.get_user", username=username), code=307)


@api.route("/users/<username>", methods=["PUT"])
def updateUser(username):
	return redirect(url_for("api.update_user", username=username), code=307)


@api.route("/users/<username>", methods=["DELETE"])
def removeUser(username):
	return redirect(url_for("api.remove_user", username=username), code=307)


@api.route("/users/<username>/password", methods=["PUT"])
def changePasswordForUser(username):
	return redirect(url_for("api.change_password_for_user", username=username), code=307)


@api.route("/users/<username>/settings", methods=["GET"])
def getSettingsForUser(username):
	return redirect(url_for("api.get_settings_for_user", username=username), code=307)


@api.route("/users/<username>/settings", methods=["PATCH"])
def changeSettingsForUser(username):
	return redirect(url_for("api.change_settings_for_user", username=username), code=307)


@api.route("/users/<username>/apikey", methods=["DELETE"])
def deleteApikeyForUser(username):
	return redirect(url_for("api.delete_apikey_for_user", username=username), code=307)


@api.route("/users/<username>/apikey", methods=["POST"])
def generateApikeyForUser(username):
	return redirect(url_for("api.generate_apikey_for_user", username=username), code=307)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
In this module resides the core data structures and logic of the plugin system. It is implemented in an OctoPrint-agnostic
way and could be extracted into a separate Python module in the future.

.. autoclass:: PluginManager
   :members:

.. autoclass:: PluginInfo
   :members:

.. autoclass:: Plugin
   :members:

.. autoclass:: RestartNeedingPlugin
   :members:

.. autoclass:: SortablePlugin
   :members:

"""

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import os
import io
from collections import defaultdict, namedtuple, OrderedDict
import logging
import fnmatch
import inspect
import sys

import pkg_resources
import pkginfo

from past.builtins import unicode

from octoprint.util import to_unicode, sv
from octoprint.util.version import is_python_compatible, get_python_version_string

try:
	from os import scandir
except ImportError:
	from scandir import scandir


if sys.version_info[0] == 2:
	# noinspection PyDeprecation
	import imp
else:
	# deprecated in Python 3.4+ and hence vendored for now
	import octoprint.vendor.imp as imp


# noinspection PyDeprecation
def _find_module(name, path=None):
	if path is not None:
		spec = imp.find_module(name, [path])
	else:
		spec = imp.find_module(name)
	return spec[1], spec


# noinspection PyDeprecation
def _load_module(name, spec):
	f, filename, details = spec
	return imp.load_module(name, f, filename, details)


EntryPointOrigin = namedtuple("EntryPointOrigin", "type, entry_point, module_name, package_name, package_version")
FolderOrigin = namedtuple("FolderOrigin", "type, folder")
ModuleOrigin = namedtuple("PackageOrigin", "type, module_name, folder")

class PluginInfo(object):
	"""
	The :class:`PluginInfo` class wraps all available information about a registered plugin.

	This includes its meta data (like name, description, version, etc) as well as the actual plugin extensions like
	implementations, hooks and helpers.

	It works on Python module objects and extracts the relevant data from those via accessing the
	:ref:`control properties <sec-plugins-controlproperties>`.

	Arguments:
	    key (str): Identifier of the plugin
	    location (str): Installation folder of the plugin
	    instance (module): Plugin module instance - this may be ``None`` if the plugin has been blacklisted!
	    name (str): Human readable name of the plugin
	    version (str): Version of the plugin
	    description (str): Description of the plugin
	    author (str): Author of the plugin
	    url (str): URL of the website of the plugin
	    license (str): License of the plugin
	"""

	attr_name = '__plugin_name__'
	""" Module attribute from which to retrieve the plugin's human readable name. """

	attr_description = '__plugin_description__'
	""" Module attribute from which to retrieve the plugin's description. """

	attr_disabling_discouraged = '__plugin_disabling_discouraged__'
	""" Module attribute from which to retrieve the reason why disabling the plugin is discouraged. Only effective if ``self.bundled`` is True. """

	attr_version = '__plugin_version__'
	""" Module attribute from which to retrieve the plugin's version. """

	attr_author = '__plugin_author__'
	""" Module attribute from which to retrieve the plugin's author. """

	attr_url = '__plugin_url__'
	""" Module attribute from which to retrieve the plugin's website URL. """

	attr_license = '__plugin_license__'
	""" Module attribute from which to retrieve the plugin's license. """

	attr_pythoncompat = '__plugin_pythoncompat__'
	""" 
	Module attribute from which to retrieve the plugin's python compatibility string. 
	
	If unset a default of ``>=2.7,<3`` will be assumed, meaning that the plugin will be considered compatible to
	Python 2 but not Python 3.
	
	To mark a plugin as Python 3 compatible, a string of ``>=2.7,<4`` is recommended.
	
	Bundled plugins will automatically be assumed to be compatible.
	"""

	attr_hidden = '__plugin_hidden__'
	"""
	Module attribute from which to determine if the plugin's hidden or not.
	
	Only evaluated for bundled plugins, in order to hide them from the Plugin Manager
	and similar places.
	"""

	attr_hooks = '__plugin_hooks__'
	""" Module attribute from which to retrieve the plugin's provided hooks. """

	attr_implementation = '__plugin_implementation__'
	""" Module attribute from which to retrieve the plugin's provided mixin implementation. """

	attr_implementations = '__plugin_implementations__'
	"""
	Module attribute from which to retrieve the plugin's provided implementations.

	This deprecated attribute will only be used if a plugin does not yet offer :attr:`attr_implementation`. Only the
	first entry will be evaluated.

	.. deprecated:: 1.2.0-dev-694

	   Use :attr:`attr_implementation` instead.
	"""

	attr_helpers = '__plugin_helpers__'
	""" Module attribute from which to retrieve the plugin's provided helpers. """

	attr_check = '__plugin_check__'
	""" Module attribute which to call to determine if the plugin can be loaded. """

	attr_init = '__plugin_init__'
	"""
	Module attribute which to call when loading the plugin.

	This deprecated attribute will only be used if a plugin does not yet offer :attr:`attr_load`.

	.. deprecated:: 1.2.0-dev-720

	   Use :attr:`attr_load` instead.
	"""

	attr_load = '__plugin_load__'
	""" Module attribute which to call when loading the plugin. """

	attr_unload = '__plugin_unload__'
	""" Module attribute which to call when unloading the plugin. """

	attr_enable = '__plugin_enable__'
	""" Module attribute which to call when enabling the plugin. """

	attr_disable = '__plugin_disable__'
	""" Module attribute which to call when disabling the plugin. """

	def __init__(self, key, location, instance, name=None, version=None, description=None, author=None, url=None, license=None, parsed_metadata=None):
		self.key = key
		self.location = location
		self.instance = instance
		self.origin = None
		self.enabled = True
		self.blacklisted = False
		self.forced_disabled = False
		self.incompatible = False
		self.bundled = False
		self.loaded = False
		self.managable = True
		self.needs_restart = False
		self.invalid_syntax = False

		self._name = name
		self._version = version
		self._description = description
		self._author = author
		self._url = url
		self._license = license

		self._logger = logging.getLogger(__name__)

		self._cached_parsed_metadata = parsed_metadata
		if self._cached_parsed_metadata is None:
			self._cached_parsed_metadata = self._parse_metadata()

	def validate(self, phase, additional_validators=None):
		result = True

		if phase == "before_import":
			result = not self.forced_disabled and not self.blacklisted and not self.incompatible and not self.invalid_syntax and result

		elif phase == "before_load":
			# if the plugin still uses __plugin_init__, log a deprecation warning and move it to __plugin_load__
			if hasattr(self.instance, self.__class__.attr_init):
				if not hasattr(self.instance, self.__class__.attr_load):
					# deprecation warning
					import warnings
					warnings.warn("{name} uses deprecated control property __plugin_init__, use __plugin_load__ instead".format(name=self.key), DeprecationWarning)

					# move it
					init = getattr(self.instance, self.__class__.attr_init)
					setattr(self.instance, self.__class__.attr_load, init)

				# delete __plugin_init__
				delattr(self.instance, self.__class__.attr_init)

		elif phase == "after_load":
			# if the plugin still uses __plugin_implementations__, log a deprecation warning and put the first
			# item into __plugin_implementation__
			if hasattr(self.instance, self.__class__.attr_implementations):
				if not hasattr(self.instance, self.__class__.attr_implementation):
					# deprecation warning
					import warnings
					warnings.warn("{name} uses deprecated control property __plugin_implementations__, use __plugin_implementation__ instead - only the first implementation of {name} will be recognized".format(name=self.key), DeprecationWarning)

					# put first item into __plugin_implementation__
					implementations = getattr(self.instance, self.__class__.attr_implementations)
					if len(implementations) > 0:
						setattr(self.instance, self.__class__.attr_implementation, implementations[0])

				# delete __plugin_implementations__
				delattr(self.instance, self.__class__.attr_implementations)

		if additional_validators is not None:
			for validator in additional_validators:
				result = validator(phase, self) and result

		return result

	def __str__(self):
		if self.version:
			return "{name} ({version})".format(name=self.name, version=self.version)
		else:
			return to_unicode(self.name)

	def long_str(self, show_bundled=False, bundled_strs=(" [B]", ""),
	             show_location=False, location_str=" - {location}",
	             show_enabled=False, enabled_strs=("* ", "  ", "X ", "C ")):
		"""
		Long string representation of the plugin's information. Will return a string of the format ``<enabled><str(self)><bundled><location>``.

		``enabled``, ``bundled`` and ``location`` will only be displayed if the corresponding flags are set to ``True``.
		The will be filled from ``enabled_str``, ``bundled_str`` and ``location_str`` as follows:

		``enabled_str``
		    a 4-tuple, the first entry being the string to insert when the plugin is enabled, the second
		    entry the string to insert when it is not, the third entry the string when it is blacklisted
		    and the fourth when it is incompatible.
		``bundled_str``
		    a 2-tuple, the first entry being the string to insert when the plugin is bundled, the second
		    entry the string to insert when it is not.
		``location_str``
		    a format string (to be parsed with ``str.format``), the ``{location}`` placeholder will be
		    replaced with the plugin's installation folder on disk.

		Arguments:
		    show_enabled (boolean): whether to show the ``enabled`` part
		    enabled_strs (tuple): the 2-tuple containing the two possible strings to use for displaying the enabled state
		    show_bundled (boolean): whether to show the ``bundled`` part
		    bundled_strs(tuple): the 2-tuple containing the two possible strings to use for displaying the bundled state
		    show_location (boolean): whether to show the ``location`` part
		    location_str (str): the format string to use for displaying the plugin's installation location

		Returns:
		    str: The long string representation of the plugin as described above
		"""
		if show_enabled:
			if self.incompatible:
				ret = to_unicode(enabled_strs[3])
			elif self.blacklisted:
				ret = to_unicode(enabled_strs[2])
			elif not self.enabled:
				ret = to_unicode(enabled_strs[1])
			else:
				ret = to_unicode(enabled_strs[0])
		else:
			ret = ""

		ret += unicode(self)

		if show_bundled:
			ret += to_unicode(bundled_strs[0]) if self.bundled else to_unicode(bundled_strs[1])

		if show_location and self.location:
			ret += to_unicode(location_str).format(location=self.location)

		return ret

	def get_hook(self, hook):
		"""
		Arguments:
		    hook (str): Hook to return.

		Returns:
		    callable or None: Handler for the requested ``hook`` or None if no handler is registered.
		"""

		if not hook in self.hooks:
			return None
		return self.hooks[hook]

	def get_implementation(self, *types):
		"""
		Arguments:
		    types (list): List of :class:`Plugin` sub classes all returned implementations need to implement.

		Returns:
		    object: The plugin's implementation if it matches all of the requested ``types``, None otherwise.
		"""

		if not self.implementation:
			return None

		for t in types:
			if not isinstance(self.implementation, t):
				return None

		return self.implementation

	@property
	def name(self):
		"""
		Human readable name of the plugin. Will be taken from name attribute of the plugin module if available,
		otherwise from the ``name`` supplied during construction with a fallback to ``key``.

		Returns:
		    str: Name of the plugin, fallback is the plugin's identifier.
		"""
		return self._get_instance_attribute(self.__class__.attr_name,
		                                    defaults=(self._name, self.key),
		                                    incl_metadata=True)

	@property
	def description(self):
		"""
		Description of the plugin. Will be taken from the description attribute of the plugin module as defined in
		:attr:`attr_description` if available, otherwise from the ``description`` supplied during construction.
		May be None.

		Returns:
		    str or None: Description of the plugin.
		"""
		return self._get_instance_attribute(self.__class__.attr_description,
		                                    default=self._description,
		                                    incl_metadata=True)

	@property
	def disabling_discouraged(self):
		"""
		Reason why disabling of this plugin is discouraged. Only evaluated for bundled plugins! Will be taken from
		the disabling_discouraged attribute of the plugin module as defined in :attr:`attr_disabling_discouraged` if
		available. False if unset or plugin not bundled.

		Returns:
		    str or None: Reason why disabling this plugin is discouraged (only for bundled plugins)
		"""
		return self._get_instance_attribute(self.__class__.attr_disabling_discouraged, default=False) if self.bundled \
			else False

	@property
	def version(self):
		"""
		Version of the plugin. Will be taken from the version attribute of the plugin module as defined in
		:attr:`attr_version` if available, otherwise from the ``version`` supplied during construction. May be None.

		Returns:
		    str or None: Version of the plugin.
		"""
		return self._version if self._version is not None else self._get_instance_attribute(self.__class__.attr_version,
		                                                                                    default=self._version,
		                                                                                    incl_metadata=True)

	@property
	def author(self):
		"""
		Author of the plugin. Will be taken from the author attribute of the plugin module as defined in
		:attr:`attr_author` if available, otherwise from the ``author`` supplied during construction. May be None.

		Returns:
		    str or None: Author of the plugin.
		"""
		return self._get_instance_attribute(self.__class__.attr_author,
		                                    default=self._author,
		                                    incl_metadata=True)

	@property
	def url(self):
		"""
		Website URL for the plugin. Will be taken from the url attribute of the plugin module as defined in
		:attr:`attr_url` if available, otherwise from the ``url`` supplied during construction. May be None.

		Returns:
		    str or None: Website URL for the plugin.
		"""
		return self._get_instance_attribute(self.__class__.attr_url,
		                                    default=self._url,
		                                    incl_metadata=True)

	@property
	def license(self):
		"""
		License of the plugin. Will be taken from the license attribute of the plugin module as defined in
		:attr:`attr_license` if available, otherwise from the ``license`` supplied during construction. May be None.

		Returns:
		    str or None: License of the plugin.
		"""
		return self._get_instance_attribute(self.__class__.attr_license,
		                                    default=self._license,
		                                    incl_metadata=True)

	@property
	def pythoncompat(self):
		"""
		Python compatibility string of the plugin module as defined in :attr:`attr_pythoncompat` if available, otherwise
		defaults to ``>=2.7,<3``.

		Returns:
		    str: Python compatibility string of the plugin
		"""
		return self._get_instance_attribute(self.__class__.attr_pythoncompat,
		                                    default=">=2.7,<3",
		                                    incl_metadata=True)

	@property
	def hidden(self):
		"""
		Hidden flag.

		Returns:
		    bool: Whether the plugin should be flagged as hidden or not
		"""
		return self._get_instance_attribute(self.__class__.attr_hidden,
		                                    default=False,
		                                    incl_metadata=True)

	@property
	def hooks(self):
		"""
		Hooks provided by the plugin. Will be taken from the hooks attribute of the plugin module as defined in
		:attr:`attr_hooks` if available, otherwise an empty dictionary is returned.

		Returns:
		    dict: Hooks provided by the plugin.
		"""
		return self._get_instance_attribute(self.__class__.attr_hooks, default={})

	@property
	def implementation(self):
		"""
		Implementation provided by the plugin. Will be taken from the implementation attribute of the plugin module
		as defined in :attr:`attr_implementation` if available, otherwise None is returned.

		Returns:
		    object: Implementation provided by the plugin.
		"""
		return self._get_instance_attribute(self.__class__.attr_implementation, default=None)

	@property
	def helpers(self):
		"""
		Helpers provided by the plugin. Will be taken from the helpers attribute of the plugin module as defined in
		:attr:`attr_helpers` if available, otherwise an empty list is returned.

		Returns:
		    dict: Helpers provided by the plugin.
		"""
		return self._get_instance_attribute(self.__class__.attr_helpers, default={})

	@property
	def check(self):
		"""
		Method for pre-load check of plugin. Will be taken from the check attribute of the plugin module as defined in
		:attr:`attr_check` if available, otherwise a lambda always returning True is returned.

		Returns:
		    callable: Check method for the plugin module which should return True if the plugin can be loaded, False
		        otherwise.
		"""
		return self._get_instance_attribute(self.__class__.attr_check, default=lambda: True)

	@property
	def load(self):
		"""
		Method for loading the plugin module. Will be taken from the load attribute of the plugin module as defined
		in :attr:`attr_load` if available, otherwise a no-operation lambda will be returned.

		Returns:
		    callable: Load method for the plugin module.
		"""
		return self._get_instance_attribute(self.__class__.attr_load, default=lambda: True)

	@property
	def unload(self):
		"""
		Method for unloading the plugin module. Will be taken from the unload attribute of the plugin module as defined
		in :attr:`attr_unload` if available, otherwise a no-operation lambda will be returned.

		Returns:
		    callable: Unload method for the plugin module.
		"""
		return self._get_instance_attribute(self.__class__.attr_unload, default=lambda: True)

	@property
	def enable(self):
		"""
		Method for enabling the plugin module. Will be taken from the enable attribute of the plugin module as defined
		in :attr:`attr_enable` if available, otherwise a no-operation lambda will be returned.

		Returns:
		    callable: Enable method for the plugin module.
		"""
		return self._get_instance_attribute(self.__class__.attr_enable, default=lambda: True)

	@property
	def disable(self):
		"""
		Method for disabling the plugin module. Will be taken from the disable attribute of the plugin module as defined
		in :attr:`attr_disable` if available, otherwise a no-operation lambda will be returned.

		Returns:
		    callable: Disable method for the plugin module.
		"""
		return self._get_instance_attribute(self.__class__.attr_disable, default=lambda: True)

	def _get_instance_attribute(self, attr, default=None, defaults=None, incl_metadata=False):
		if self.instance is None or not hasattr(self.instance, attr):
			if incl_metadata and attr in self.parsed_metadata:
				return self.parsed_metadata[attr]

			elif defaults is not None:
				for value in defaults:
					if callable(value):
						value = value()
					if value is not None:
						return value

			return default

		return getattr(self.instance, attr)

	@property
	def parsed_metadata(self):
		return self._cached_parsed_metadata

	def _parse_metadata(self):
		result = dict()

		path = self.location
		if not path:
			return result

		if os.path.isdir(path):
			path = os.path.join(self.location, "__init__.py")

		if not os.path.isfile(path):
			return result

		if not path.endswith(".py"):
			# we only support parsing plain text source files
			return result

		self._logger.debug("Parsing plugin metadata for {} from AST of {}".format(self.key, path))

		try:
			import ast

			with io.open(path, 'rb') as f:
				root = ast.parse(f.read(), filename=path)

			assignments = list(filter(lambda x: isinstance(x, ast.Assign) and x.targets,
			                          root.body))

			def extract_target_ids(node):
				return list(map(lambda x: x.id,
				                filter(lambda x: isinstance(x, ast.Name), node.targets)))

			for key in (self.__class__.attr_name, self.__class__.attr_version, self.__class__.attr_author,
			            self.__class__.attr_description, self.__class__.attr_url, self.__class__.attr_license,
			            self.__class__.attr_pythoncompat):
				for a in reversed(assignments):
					targets = extract_target_ids(a)
					if key in targets:
						if isinstance(a.value, ast.Str):
							result[key] = a.value.s

						elif isinstance(a.value, ast.Call) and hasattr(a.value, "func") \
								and a.value.func.id == "gettext" and a.value.args \
								and isinstance(a.value.args[0], ast.Str):
							result[key] = a.value.args[0].s

						break

			for key in (self.__class__.attr_hidden,):
				for a in reversed(assignments):
					targets = extract_target_ids(a)
					if key in targets:
						if isinstance(a.value, ast.Name) and a.value.id in ("True", "False"):
							result[key] = bool(a.value.id)

						break

		except SyntaxError:
			self._logger.exception("Invalid syntax in {} for plugin {}".format(path, self.key))
			self.invalid_syntax = True
		except Exception:
			self._logger.exception("Error while parsing AST from {} for plugin {}".format(path, self.key))

		return result


class PluginManager(object):
	"""
	The :class:`PluginManager` is the central component for finding, loading and accessing plugins provided to the
	system.

	It is able to discover plugins both through possible file system locations as well as customizable entry points.
	"""

	def __init__(self, plugin_folders, plugin_bases, plugin_entry_points, logging_prefix=None,
	             plugin_disabled_list=None, plugin_blacklist=None, plugin_restart_needing_hooks=None,
	             plugin_obsolete_hooks=None, plugin_validators=None, compatibility_ignored_list=None):
		self.logger = logging.getLogger(__name__)

		if logging_prefix is None:
			logging_prefix = ""
		if plugin_folders is None:
			plugin_folders = []
		if plugin_bases is None:
			plugin_bases = []
		if plugin_entry_points is None:
			plugin_entry_points = []
		if plugin_disabled_list is None:
			plugin_disabled_list = []
		if plugin_blacklist is None:
			plugin_blacklist = []
		if compatibility_ignored_list is None:
			compatibility_ignored_list = []

		self.plugin_folders = plugin_folders
		self.plugin_bases = plugin_bases
		self.plugin_entry_points = plugin_entry_points
		self.plugin_disabled_list = plugin_disabled_list
		self.plugin_blacklist = plugin_blacklist
		self.plugin_restart_needing_hooks = plugin_restart_needing_hooks
		self.plugin_obsolete_hooks = plugin_obsolete_hooks
		self.plugin_validators = plugin_validators
		self.logging_prefix = logging_prefix
		self.compatibility_ignored_list = compatibility_ignored_list

		self.enabled_plugins = dict()
		self.disabled_plugins = dict()
		self.plugin_implementations = dict()
		self.plugin_implementations_by_type = defaultdict(list)

		self._plugin_hooks = defaultdict(list)

		self.implementation_injects = dict()
		self.implementation_inject_factories = []
		self.implementation_pre_inits = []
		self.implementation_post_inits = []

		self.on_plugin_loaded = lambda *args, **kwargs: None
		self.on_plugin_unloaded = lambda *args, **kwargs: None
		self.on_plugin_enabled = lambda *args, **kwargs: None
		self.on_plugin_disabled = lambda *args, **kwargs: None
		self.on_plugin_implementations_initialized = lambda *args, **kwargs: None

		self.on_plugins_loaded = lambda *args, **kwargs: None
		self.on_plugins_enabled = lambda *args, **kwargs: None

		self.registered_clients = []

		self.marked_plugins = defaultdict(list)

		self._python_install_dir = None
		self._python_virtual_env = False
		self._detect_python_environment()

	def _detect_python_environment(self):
		from distutils.command.install import install as cmd_install
		from distutils.dist import Distribution
		import sys

		cmd = cmd_install(Distribution())
		cmd.finalize_options()

		self._python_install_dir = cmd.install_lib
		self._python_prefix = os.path.realpath(sys.prefix)
		self._python_virtual_env = hasattr(sys, "real_prefix") \
		                           or (hasattr(sys, "base_prefix") and os.path.realpath(sys.prefix) != os.path.realpath(sys.base_prefix))

	@property
	def plugins(self):
		plugins = dict(self.enabled_plugins)
		plugins.update(self.disabled_plugins)
		return plugins

	@property
	def plugin_hooks(self):
		return {key: list(map(lambda v: (v[1], v[2]), value)) for key, value in self._plugin_hooks.items()}

	def find_plugins(self, existing=None, ignore_uninstalled=True, incl_all_found=False):
		added, found = self._find_plugins(existing=existing, ignore_uninstalled=ignore_uninstalled)
		if incl_all_found:
			return added, found
		else:
			return added

	def _find_plugins(self, existing=None, ignore_uninstalled=True):
		if existing is None:
			existing = dict(self.plugins)

		result_added = OrderedDict()
		result_found = []

		if self.plugin_folders:
			try:
				added, found = self._find_plugins_from_folders(self.plugin_folders,
				                                                 existing,
				                                                 ignored_uninstalled=ignore_uninstalled)
				result_added.update(added)
				result_found += found
			except Exception:
				self.logger.exception("Error fetching plugins from folders")

		if self.plugin_entry_points:
			existing.update(result_added)
			try:
				added, found = self._find_plugins_from_entry_points(self.plugin_entry_points,
				                                                      existing,
				                                                      ignore_uninstalled=ignore_uninstalled)
				result_added.update(added)
				result_found += found
			except Exception:
				self.logger.exception("Error fetching plugins from entry points")

		return result_added, result_found

	def _find_plugins_from_folders(self, folders, existing, ignored_uninstalled=True):
		added = OrderedDict()
		found = []

		for folder in folders:
			try:
				flagged_readonly = False
				package = None
				if isinstance(folder, (list, tuple)):
					if len(folder) == 2:
						folder, flagged_readonly = folder
					elif len(folder) == 3:
						folder, package, flagged_readonly = folder
					else:
						continue
				actual_readonly = not os.access(folder, os.W_OK)

				if not os.path.exists(folder):
					self.logger.warning("Plugin folder {folder} could not be found, skipping it".format(folder=folder))
					continue

				for entry in scandir(folder):
					try:
						if entry.is_dir():
							init_py = os.path.join(entry.path, "__init__.py")
							init_pyc = os.path.join(entry.path, "__init__.pyc")

							if not os.path.isfile(init_py) and not os.path.isfile(init_pyc):
								# neither does exist, we ignore this
								continue

							key = entry.name

						elif entry.is_file():
							key, ext = os.path.splitext(entry.name)
							if ext not in (".py", ".pyc") or key.startswith("__"):
								# neither py nor pyc, or starts with __ (like __init__), we ignore this
								continue

						else:
							# whatever this is, we ignore it
							continue

						found.append(key)
						if key in existing or key in added or (ignored_uninstalled and key in self.marked_plugins["uninstalled"]):
							# plugin is already defined, ignore it
							continue

						bundled = flagged_readonly

						module_name = None
						if package:
							module_name = "{}.{}".format(package, key)

						plugin = self._import_plugin_from_module(key, module_name=module_name, folder=folder, bundled=bundled)
						if plugin:
							if module_name:
								plugin.origin = ModuleOrigin("module", module_name, folder)
							else:
								plugin.origin = FolderOrigin("folder", folder)
							plugin.managable = not flagged_readonly and not actual_readonly
							plugin.enabled = False
							added[key] = plugin
					except Exception:
						self.logger.exception("Error processing folder entry {!r} from folder {}".format(entry, folder))
			except Exception:
				self.logger.exception("Error processing folder {}".format(folder))

		return added, found

	def _find_plugins_from_entry_points(self, groups, existing, ignore_uninstalled=True):
		added = OrderedDict()
		found = []

		# let's make sure we have a current working set ...
		working_set = pkg_resources.WorkingSet()

		# ... including the user's site packages
		import site
		import sys
		if site.ENABLE_USER_SITE:
			if not site.USER_SITE in working_set.entries:
				working_set.add_entry(site.USER_SITE)
			if not site.USER_SITE in sys.path:
				site.addsitedir(site.USER_SITE)

		if not isinstance(groups, (list, tuple)):
			groups = [groups]

		def wrapped(gen):
			# to protect against some issues in installed packages that make iteration over entry points
			# fall on its face - e.g. https://groups.google.com/forum/#!msg/octoprint/DyXdqhR0U7c/kKMUsMmIBgAJ
			for entry in gen:
				try:
					yield entry
				except Exception:
					self.logger.exception("Something went wrong while processing the entry points of a package in the "
					                      "Python environment - broken entry_points.txt in some package?")

		for group in groups:
			for entry_point in wrapped(working_set.iter_entry_points(group=group, name=None)):
				try:
					key = entry_point.name
					module_name = entry_point.module_name
					version = entry_point.dist.version

					found.append(key)
					if key in existing or key in added or (ignore_uninstalled and key in self.marked_plugins["uninstalled"]):
						# plugin is already defined or marked as uninstalled, ignore it
						continue

					kwargs = dict(module_name=module_name, version=version)
					package_name = entry_point.dist.project_name
					try:
						entry_point_metadata = EntryPointMetadata(entry_point)
					except Exception:
						self.logger.exception("Something went wrong while retrieving metadata for module {}".format(module_name))
					else:
						kwargs.update(dict(
							name=entry_point_metadata.name,
							summary=entry_point_metadata.summary,
							author=entry_point_metadata.author,
							url=entry_point_metadata.home_page,
							license=entry_point_metadata.license
						))

					plugin = self._import_plugin_from_module(key, **kwargs)
					if plugin:
						plugin.origin = EntryPointOrigin("entry_point", group, module_name, package_name, version)
						plugin.enabled = False

						# plugin is manageable if its location is writable and OctoPrint
						# is either not running from a virtual env or the plugin is
						# installed in that virtual env - the virtual env's pip will not
						# allow us to uninstall stuff that is installed outside
						# of the virtual env, so this check is necessary
						plugin.managable = os.access(plugin.location, os.W_OK) \
						                   and (not self._python_virtual_env
						                        or is_sub_path_of(plugin.location, self._python_prefix)
						                        or is_editable_install(self._python_install_dir,
						                                               package_name,
						                                               module_name,
						                                               plugin.location))

						added[key] = plugin
				except Exception:
					self.logger.exception("Error processing entry point {!r} for group {}".format(entry_point, group))

		return added, found

	def _import_plugin_from_module(self, key, folder=None, module_name=None, name=None, version=None, summary=None,
	                               author=None, url=None, license=None, bundled=False):
		# TODO error handling
		try:
			if folder:
				location, spec = _find_module(key, path=folder)
			elif module_name:
				location, spec = _find_module(module_name)
			else:
				return None
		except Exception:
			self.logger.exception("Could not locate plugin {key}".format(key=key))
			return None

		# Create a simple dummy entry first ...
		plugin = PluginInfo(key, location, None, name=name, version=version, description=summary, author=author,
		                    url=url, license=license)
		plugin.bundled = bundled

		if self._is_plugin_disabled(key):
			self.logger.info("Plugin {} is disabled.".format(plugin))
			plugin.forced_disabled = True

		if self._is_plugin_blacklisted(key) or (plugin.version is not None and self._is_plugin_version_blacklisted(key, plugin.version)):
			self.logger.warning("Plugin {} is blacklisted.".format(plugin))
			plugin.blacklisted = True

		python_version = get_python_version_string()
		if self._is_plugin_incompatible(key, plugin):
			if plugin.invalid_syntax:
				self.logger.warning("Plugin {} can't be compiled under Python {} due to invalid syntax".format(plugin, python_version))
			else:
				self.logger.warning("Plugin {} is not compatible to Python {} (compatibility string: {}).".format(plugin, python_version, plugin.pythoncompat))
			plugin.incompatible = True

		if not plugin.validate("before_import", additional_validators=self.plugin_validators):
			return plugin

		# ... then create and return the real one
		return self._import_plugin(key, spec,
		                           module_name=module_name, name=name, location=plugin.location, version=version,
		                           summary=summary, author=author, url=url, license=license, bundled=bundled,
		                           parsed_metadata=plugin.parsed_metadata)

	def _import_plugin(self, key, spec, module_name=None, name=None, location=None, version=None, summary=None, author=None, url=None, license=None, bundled=False, parsed_metadata=None):
		try:
			if module_name:
				module = _load_module(module_name, spec)
			else:
				module = _load_module(key, spec)

			plugin = PluginInfo(key, location, module,
			                    name=name,
			                    version=version,
			                    description=summary,
			                    author=author,
			                    url=url,
			                    license=license,
			                    parsed_metadata=parsed_metadata)

			plugin.bundled = bundled
		except Exception:
			self.logger.exception("Error loading plugin {key}".format(key=key))
			return None

		if plugin.check():
			return plugin
		else:
			self.logger.info("Plugin {plugin} did not pass check, not loading.".format(plugin=str(plugin)))
			return None

	def _is_plugin_disabled(self, key):
		return key in self.plugin_disabled_list or key.endswith('disabled')

	def _is_plugin_blacklisted(self, key):
		return key in self.plugin_blacklist

	def _is_plugin_incompatible(self, key, plugin):
		return not plugin.bundled \
		       and not is_python_compatible(plugin.pythoncompat) \
		       and key not in self.compatibility_ignored_list

	def _is_plugin_version_blacklisted(self, key, version):
		def matches_plugin(entry):
			if isinstance(entry, (tuple, list)) and len(entry) == 2:
				entry_key, entry_version = entry
				return entry_key == key and entry_version == version
			return False

		return any(map(lambda entry: matches_plugin(entry),
		               self.plugin_blacklist))

	def reload_plugins(self, startup=False, initialize_implementations=True, force_reload=None):
		self.logger.info("Loading plugins from {folders} and installed plugin packages...".format(
			folders=", ".join(map(lambda x: x[0] if isinstance(x, tuple) else str(x), self.plugin_folders))
		))

		if force_reload is None:
			force_reload = []

		added, found = self.find_plugins(existing=dict((k, v) for k, v in self.plugins.items() if not k in force_reload),
		                                 incl_all_found=True)

		# let's clean everything we DIDN'T find first
		removed = [key for key in list(self.enabled_plugins.keys()) + list(self.disabled_plugins.keys()) if key not in found]
		for key in removed:
			try:
				del self.enabled_plugins[key]
			except KeyError:
				pass

			try:
				del self.disabled_plugins[key]
			except KeyError:
				pass

		self.disabled_plugins.update(added)

		# 1st pass: loading the plugins
		for name, plugin in added.items():
			try:
				if not plugin.blacklisted and not plugin.forced_disabled and not plugin.incompatible:
					self.load_plugin(name, plugin, startup=startup, initialize_implementation=initialize_implementations)
			except PluginNeedsRestart:
				pass
			except PluginLifecycleException as e:
				self.logger.info(str(e))

		self.on_plugins_loaded(startup=startup,
							   initialize_implementations=initialize_implementations,
							   force_reload=force_reload)

		# 2nd pass: enabling those plugins that need enabling
		for name, plugin in added.items():
			try:
				if plugin.loaded and not plugin.forced_disabled and not plugin.incompatible:
					if plugin.blacklisted:
						self.logger.warning("Plugin {} is blacklisted. Not enabling it.".format(plugin))
						continue
					self.enable_plugin(name, plugin=plugin, initialize_implementation=initialize_implementations, startup=startup)
			except PluginNeedsRestart:
				pass
			except PluginLifecycleException as e:
				self.logger.info(str(e))

		self.on_plugins_enabled(startup=startup,
								initialize_implementations=initialize_implementations,
								force_reload=force_reload)

		if len(self.enabled_plugins) <= 0:
			self.logger.info("No plugins found")
		else:
			self.logger.info("Found {count} plugin(s) providing {implementations} mixin implementations, {hooks} hook handlers".format(
				count=len(self.enabled_plugins) + len(self.disabled_plugins),
				implementations=len(self.plugin_implementations),
				hooks=sum(map(lambda x: len(x), self.plugin_hooks.values()))
			))

	def mark_plugin(self, name, **kwargs):
		if not name in self.plugins:
			self.logger.debug("Trying to mark an unknown plugin {name}".format(**locals()))

		for key, value in kwargs.items():
			if value is None:
				continue

			if value and not name in self.marked_plugins[key]:
				self.marked_plugins[key].append(name)
			elif not value and name in self.marked_plugins[key]:
				self.marked_plugins[key].remove(name)

	def is_plugin_marked(self, name, key):
		if not name in self.plugins:
			return False

		return name in self.marked_plugins[key]

	def load_plugin(self, name, plugin=None, startup=False, initialize_implementation=True):
		if not name in self.plugins:
			self.logger.warning("Trying to load an unknown plugin {name}".format(**locals()))
			return

		if plugin is None:
			plugin = self.plugins[name]

		try:
			if not plugin.validate("before_load", additional_validators=self.plugin_validators):
				return

			plugin.load()
			plugin.validate("after_load", additional_validators=self.plugin_validators)
			self.on_plugin_loaded(name, plugin)
			plugin.loaded = True

			self.logger.debug("Loaded plugin {name}: {plugin}".format(**locals()))
		except PluginLifecycleException as e:
			raise e
		except Exception:
			self.logger.exception("There was an error loading plugin %s" % name)

	def unload_plugin(self, name):
		if not name in self.plugins:
			self.logger.warning("Trying to unload unknown plugin {name}".format(**locals()))
			return

		plugin = self.plugins[name]

		try:
			if plugin.enabled:
				self.disable_plugin(name, plugin=plugin)

			plugin.unload()
			self.on_plugin_unloaded(name, plugin)

			if name in self.enabled_plugins:
				del self.enabled_plugins[name]

			if name in self.disabled_plugins:
				del self.disabled_plugins[name]

			plugin.loaded = False

			self.logger.debug("Unloaded plugin {name}: {plugin}".format(**locals()))
		except PluginLifecycleException as e:
			raise e
		except Exception:
			self.logger.exception("There was an error unloading plugin {name}".format(**locals()))

			# make sure the plugin is NOT in the list of enabled plugins but in the list of disabled plugins
			if name in self.enabled_plugins:
				del self.enabled_plugins[name]
			if not name in self.disabled_plugins:
				self.disabled_plugins[name] = plugin

	def enable_plugin(self, name, plugin=None, initialize_implementation=True, startup=False):
		if not name in self.disabled_plugins:
			self.logger.warning("Tried to enable plugin {name}, however it is not disabled".format(**locals()))
			return

		if plugin is None:
			plugin = self.disabled_plugins[name]

		if not startup and self.is_restart_needing_plugin(plugin):
			raise PluginNeedsRestart(name)

		if self.has_obsolete_hooks(plugin):
			raise PluginCantEnable(name, "Dependency on obsolete hooks detected, full functionality cannot be guaranteed")

		try:
			if not plugin.validate("before_enable", additional_validators=self.plugin_validators):
				return False

			plugin.enable()
			self._activate_plugin(name, plugin)
		except PluginLifecycleException as e:
			raise e
		except Exception:
			self.logger.exception("There was an error while enabling plugin {name}".format(**locals()))
			return False
		else:
			if name in self.disabled_plugins:
				del self.disabled_plugins[name]
			self.enabled_plugins[name] = plugin
			plugin.enabled = True

			if plugin.implementation:
				if initialize_implementation:
					if not self.initialize_implementation_of_plugin(name, plugin):
						return False
				plugin.implementation.on_plugin_enabled()
			self.on_plugin_enabled(name, plugin)

			self.logger.debug("Enabled plugin {name}: {plugin}".format(**locals()))

		return True

	def disable_plugin(self, name, plugin=None):
		if not name in self.enabled_plugins:
			self.logger.warning("Tried to disable plugin {name}, however it is not enabled".format(**locals()))
			return

		if plugin is None:
			plugin = self.enabled_plugins[name]

		if self.is_restart_needing_plugin(plugin):
			raise PluginNeedsRestart(name)

		try:
			plugin.disable()
			self._deactivate_plugin(name, plugin)
		except PluginLifecycleException as e:
			raise e
		except Exception:
			self.logger.exception("There was an error while disabling plugin {name}".format(**locals()))
			return False
		else:
			if name in self.enabled_plugins:
				del self.enabled_plugins[name]
			self.disabled_plugins[name] = plugin
			plugin.enabled = False

			if plugin.implementation:
				plugin.implementation.on_plugin_disabled()
			self.on_plugin_disabled(name, plugin)

			self.logger.debug("Disabled plugin {name}: {plugin}".format(**locals()))

		return True

	def _activate_plugin(self, name, plugin):
		plugin.hotchangeable = self.is_restart_needing_plugin(plugin)

		# evaluate registered hooks
		for hook, definition in plugin.hooks.items():
			try:
				callback, order = self._get_callback_and_order(definition)
			except ValueError as e:
				self.logger.warning("There is something wrong with the hook definition {} for plugin {}: {}".format(definition, name, str(e)))
				continue

			self._plugin_hooks[hook].append((order, name, callback))
			self._sort_hooks(hook)

		# evaluate registered implementation
		if plugin.implementation:
			mixins = self.mixins_matching_bases(plugin.implementation.__class__, *self.plugin_bases)
			for mixin in mixins:
				self.plugin_implementations_by_type[mixin].append((name, plugin.implementation))

			self.plugin_implementations[name] = plugin.implementation

	def _deactivate_plugin(self, name, plugin):
		for hook, definition in plugin.hooks.items():
			try:
				callback, order = self._get_callback_and_order(definition)
			except ValueError as e:
				self.logger.warning("There is something wrong with the hook definition {} for plugin {}: {}".format(definition, name, str(e)))
				continue

			try:
				self._plugin_hooks[hook].remove((order, name, callback))
				self._sort_hooks(hook)
			except ValueError:
				# that's ok, the plugin was just not registered for the hook
				pass

		if plugin.implementation is not None:
			if name in self.plugin_implementations:
				del self.plugin_implementations[name]

			mixins = self.mixins_matching_bases(plugin.implementation.__class__, *self.plugin_bases)
			for mixin in mixins:
				try:
					self.plugin_implementations_by_type[mixin].remove((name, plugin.implementation))
				except ValueError:
					# that's ok, the plugin was just not registered for the type
					pass

	def is_restart_needing_plugin(self, plugin):
		return plugin.needs_restart or self.has_restart_needing_implementation(plugin) or self.has_restart_needing_hooks(plugin)

	def has_restart_needing_implementation(self, plugin):
		return self.has_any_of_mixins(plugin, RestartNeedingPlugin)

	def has_restart_needing_hooks(self, plugin):
		return self.has_any_of_hooks(plugin, self.plugin_restart_needing_hooks)

	def has_obsolete_hooks(self, plugin):
		return self.has_any_of_hooks(plugin, self.plugin_obsolete_hooks)

	def is_restart_needing_hook(self, hook):
		return self.hook_matches_hooks(hook, self.plugin_restart_needing_hooks)

	def is_obsolete_hook(self, hook):
		return self.hook_matches_hooks(hook, self.plugin_obsolete_hooks)

	@staticmethod
	def has_any_of_hooks(plugin, *hooks):
		"""
		Tests if the ``plugin`` contains any of the provided ``hooks``.

		Uses :func:`octoprint.plugin.core.PluginManager.hook_matches_hooks`.

		Args:
			plugin: plugin to test hooks for
			*hooks: hooks to test against

		Returns:
			(bool): True if any of the plugin's hooks match the provided hooks,
			        False otherwise.
		"""

		if hooks and len(hooks) == 1 and isinstance(hooks[0], (list, tuple)):
			hooks = hooks[0]

		hooks = list(filter(lambda hook: hook is not None, hooks))
		if not hooks:
			return False
		if not plugin or not plugin.hooks:
			return False

		plugin_hooks = plugin.hooks.keys()

		return any(map(lambda hook: PluginManager.hook_matches_hooks(hook, *hooks),
		               plugin_hooks))

	@staticmethod
	def hook_matches_hooks(hook, *hooks):
		"""
		Tests if ``hook`` matches any of the provided ``hooks`` to test for.

		``hook`` is expected to be an exact hook name.

		``hooks`` is expected to be a list containing one or more hook names or
		patterns. That can be either an exact hook name or an
		:func:`fnmatch.fnmatch` pattern.

		Args:
			hook: the hook to test
			hooks: the hook name patterns to test against

		Returns:
			(bool): True if the ``hook`` matches any of the ``hooks``, False otherwise.

		"""

		if hooks and len(hooks) == 1 and isinstance(hooks[0], (list, tuple)):
			hooks = hooks[0]

		hooks = list(filter(lambda hook: hook is not None, hooks))
		if not hooks:
			return False
		if not hook:
			return False

		return any(map(lambda h: fnmatch.fnmatch(hook, h),
		               hooks))

	@staticmethod
	def mixins_matching_bases(klass, *bases):
		result = set()
		for c in inspect.getmro(klass):
			if c == klass or c in bases:
				# ignore the exact class and our bases
				continue
			if issubclass(c, bases):
				result.add(c)
		return result

	@staticmethod
	def has_any_of_mixins(plugin, *mixins):
		"""
		Tests if the ``plugin`` has an implementation implementing any
		of the provided ``mixins``.

		Args:
			plugin: plugin for which to check the implementation
			*mixins: mixins to test against

		Returns:
			(bool): True if the plugin's implementation implements any of the
			        provided mixins, False otherwise.
		"""

		if mixins and len(mixins) == 1 and isinstance(mixins[0], (list, tuple)):
			mixins = mixins[0]

		mixins = list(filter(lambda mixin: mixin is not None, mixins))
		if not mixins:
			return False
		if not plugin or not plugin.implementation:
			return False

		return isinstance(plugin.implementation, tuple(mixins))

	def initialize_implementations(self, additional_injects=None, additional_inject_factories=None, additional_pre_inits=None, additional_post_inits=None):
		for name, plugin in self.enabled_plugins.items():
			self.initialize_implementation_of_plugin(name, plugin,
			                                         additional_injects=additional_injects,
			                                         additional_inject_factories=additional_inject_factories,
			                                         additional_pre_inits=additional_pre_inits,
			                                         additional_post_inits=additional_post_inits)

		self.logger.info("Initialized {count} plugin implementation(s)".format(count=len(self.plugin_implementations)))

	def initialize_implementation_of_plugin(self, name, plugin, additional_injects=None, additional_inject_factories=None, additional_pre_inits=None, additional_post_inits=None):
		if plugin.implementation is None:
			return

		return self.initialize_implementation(name, plugin, plugin.implementation,
		                               additional_injects=additional_injects,
		                               additional_inject_factories=additional_inject_factories,
		                               additional_pre_inits=additional_pre_inits,
		                               additional_post_inits=additional_post_inits)

	def initialize_implementation(self, name, plugin, implementation, additional_injects=None, additional_inject_factories=None, additional_pre_inits=None, additional_post_inits=None):
		if additional_injects is None:
			additional_injects = dict()
		if additional_inject_factories is None:
			additional_inject_factories = []
		if additional_pre_inits is None:
			additional_pre_inits = []
		if additional_post_inits is None:
			additional_post_inits = []

		injects = self.implementation_injects
		injects.update(additional_injects)

		inject_factories = self.implementation_inject_factories
		inject_factories += additional_inject_factories

		pre_inits = self.implementation_pre_inits
		pre_inits += additional_pre_inits

		post_inits = self.implementation_post_inits
		post_inits += additional_post_inits

		try:
			kwargs = dict(injects)

			kwargs.update(dict(
				identifier=name,
				plugin_name=plugin.name,
				plugin_version=plugin.version,
				plugin_info=plugin,
				basefolder=os.path.realpath(plugin.location),
				logger=logging.getLogger(self.logging_prefix + name),
				))

			# inject the additional_injects
			for arg, value in kwargs.items():
				setattr(implementation, "_" + arg, value)

			# inject any injects produced in the additional_inject_factories
			for factory in inject_factories:
				try:
					return_value = factory(name, implementation)
				except Exception:
					self.logger.exception("Exception while executing injection factory %r" % factory)
				else:
					if return_value is not None:
						if isinstance(return_value, dict):
							for arg, value in return_value.items():
								setattr(implementation, "_" + arg, value)

			# execute any additional pre init methods
			for pre_init in pre_inits:
				pre_init(name, implementation)

			implementation.initialize()

			# execute any additional post init methods
			for post_init in post_inits:
				post_init(name, implementation)

		except Exception as e:
			self._deactivate_plugin(name, plugin)
			plugin.enabled = False

			if isinstance(e, PluginLifecycleException):
				raise e
			else:
				self.logger.exception("Exception while initializing plugin {name}, disabling it".format(**locals()))
				return False
		else:
			self.on_plugin_implementations_initialized(name, plugin)

		self.logger.debug("Initialized plugin mixin implementation for plugin {name}".format(**locals()))
		return True


	def log_all_plugins(self, show_bundled=True, bundled_str=(" (bundled)", ""), show_location=True,
	                    location_str=" = {location}", show_enabled=True, enabled_str=(" ", "!", "#", "*"),
	                    only_to_handler=None):
		all_plugins = list(self.enabled_plugins.values()) + list(self.disabled_plugins.values())

		def _log(message, level=logging.INFO):
			if only_to_handler is not None:
				import octoprint.logging
				octoprint.logging.log_to_handler(self.logger, only_to_handler, level, message, [])
			else:
				self.logger.log(level, message)

		if len(all_plugins) <= 0:
			_log("No plugins available")
		else:
			formatted_plugins = "\n".join(map(lambda x: "| " + x.long_str(show_bundled=show_bundled,
				                                                          bundled_strs=bundled_str,
				                                                          show_location=show_location,
				                                                          location_str=location_str,
				                                                          show_enabled=show_enabled,
				                                                          enabled_strs=enabled_str),
				                              sorted(self.plugins.values(), key=lambda x: unicode(x).lower())))
			legend = "Prefix legend: {1} = disabled, {2} = blacklisted, {3} = incompatible".format(*enabled_str)
			_log("{count} plugin(s) registered with the system:\n{plugins}\n{legend}".format(count=len(all_plugins),
			                                                                                 plugins=formatted_plugins,
			                                                                                 legend=legend))

	def get_plugin(self, identifier, require_enabled=True):
		"""
		Retrieves the module of the plugin identified by ``identifier``. If the plugin is not registered or disabled and
		``required_enabled`` is True (the default) None will be returned.

		Arguments:
		    identifier (str): The identifier of the plugin to retrieve.
		    require_enabled (boolean): Whether to only return the plugin if is enabled (True, default) or also if it's
		        disabled.

		Returns:
		    module: The requested plugin module or None
		"""

		plugin_info = self.get_plugin_info(identifier, require_enabled=require_enabled)
		if plugin_info is not None:
			return plugin_info.instance
		return None

	def get_plugin_info(self, identifier, require_enabled=True):
		"""
		Retrieves the :class:`PluginInfo` instance identified by ``identifier``. If the plugin is not registered or
		disabled and ``required_enabled`` is True (the default) None will be returned.

		Arguments:
		    identifier (str): The identifier of the plugin to retrieve.
		    require_enabled (boolean): Whether to only return the plugin if is enabled (True, default) or also if it's
		        disabled.

		Returns:
		    ~.PluginInfo: The requested :class:`PluginInfo` or None
		"""

		if identifier in self.enabled_plugins:
			return self.enabled_plugins[identifier]
		elif not require_enabled and identifier in self.disabled_plugins:
			return self.disabled_plugins[identifier]

		return None

	def get_hooks(self, hook):
		"""
		Retrieves all registered handlers for the specified hook.

		Arguments:
		    hook (str): The hook for which to retrieve the handlers.

		Returns:
		    dict: A dict containing all registered handlers mapped by their plugin's identifier.
		"""

		if not hook in self.plugin_hooks:
			return dict()

		result = OrderedDict()
		for h in self.plugin_hooks[hook]:
			result[h[0]] = h[1]
		return result

	def get_implementations(self, *types, **kwargs):
		"""
		Get all mixin implementations that implement *all* of the provided ``types``.

		Arguments:
		    types (one or more type): The types a mixin implementation needs to implement in order to be returned.

		Returns:
		    list: A list of all found implementations
		"""

		sorting_context = kwargs.get("sorting_context", None)

		result = None

		for t in types:
			implementations = self.plugin_implementations_by_type[t]
			if result is None:
				result = set(implementations)
			else:
				result = result.intersection(implementations)

		if result is None:
			return []

		def sort_func(impl):
			sorting_value = None
			if sorting_context is not None and isinstance(impl[1], SortablePlugin):
				try:
					sorting_value = impl[1].get_sorting_key(sorting_context)
				except Exception:
					self.logger.exception("Error while trying to retrieve sorting order for plugin {}".format(impl[0]))

				if sorting_value is not None:
					try:
						sorting_value = int(sorting_value)
					except ValueError:
						self.logger.warning("The order value returned by {} for sorting context {} is not a valid integer, ignoring it".format(impl[0], sorting_context))
						sorting_value = None

			plugin_info = self.get_plugin_info(impl[0], require_enabled=False)
			return sorting_value is None, sv(sorting_value), not plugin_info.bundled if plugin_info else True, sv(impl[0])

		return [impl[1] for impl in sorted(result, key=sort_func)]

	def get_filtered_implementations(self, f, *types, **kwargs):
		"""
		Get all mixin implementations that implement *all* of the provided ``types`` and match the provided filter `f`.

		Arguments:
		    f (callable): A filter function returning True for implementations to return and False for those to exclude.
		    types (one or more type): The types a mixin implementation needs to implement in order to be returned.

		Returns:
		    list: A list of all found and matching implementations.
		"""

		assert callable(f)
		implementations = self.get_implementations(*types, sorting_context=kwargs.get("sorting_context", None))
		return list(filter(f, implementations))

	def get_helpers(self, name, *helpers):
		"""
		Retrieves the named ``helpers`` for the plugin with identifier ``name``.

		If the plugin is not available, returns None. Otherwise returns a :class:`dict` with the requested plugin
		helper names mapped to the method - if a helper could not be resolved, it will be missing from the dict.

		Arguments:
		    name (str): Identifier of the plugin for which to look up the ``helpers``.
		    helpers (one or more str): Identifiers of the helpers of plugin ``name`` to return.

		Returns:
		    dict: A dictionary of all resolved helpers, mapped by their identifiers, or None if the plugin was not
		        registered with the system.
		"""

		if not name in self.enabled_plugins:
			return None
		plugin = self.enabled_plugins[name]

		all_helpers = plugin.helpers
		if len(helpers):
			return dict((k, v) for (k, v) in all_helpers.items() if k in helpers)
		else:
			return all_helpers

	def register_message_receiver(self, client):
		"""
		Registers a ``client`` for receiving plugin messages. The ``client`` needs to be a callable accepting two
		input arguments, ``plugin`` (the sending plugin's identifier) and ``data`` (the message itself), and one
		optional keyword argument, ``permissions`` (an optional list of permissions to test against).
		"""

		if client is None:
			return
		self.registered_clients.append(client)

	def unregister_message_receiver(self, client):
		"""
		Unregisters a ``client`` for receiving plugin messages.
		"""

		try:
			self.registered_clients.remove(client)
		except ValueError:
			# not registered
			pass

	def send_plugin_message(self, plugin, data, permissions=None):
		"""
		Sends ``data`` in the name of ``plugin`` to all currently registered message receivers by invoking them
		with the three arguments.

		Arguments:
		    plugin (str): The sending plugin's identifier.
		    data (object): The message.
		    permissions (list): A list of permissions to test against in the client.
		"""

		for client in self.registered_clients:
			try:
				client(plugin, data, permissions=permissions)
			except Exception:
				self.logger.exception("Exception while sending plugin data to client")

	def _sort_hooks(self, hook):
		self._plugin_hooks[hook] = sorted(self._plugin_hooks[hook],
		                                  key=lambda x: (x[0] is None, sv(x[0]), sv(x[1]), sv(x[2])))

	def _get_callback_and_order(self, hook):
		if callable(hook):
			return hook, None

		elif isinstance(hook, tuple) and len(hook) == 2:
			callback, order = hook

			# test that callback is a callable
			if not callable(callback):
				raise ValueError("Hook callback is not a callable")

			# test that number is an int
			try:
				int(order)
			except ValueError:
				raise ValueError("Hook order is not a number")

			return callback, order

		else:
			raise ValueError("Invalid hook definition, neither a callable nor a 2-tuple (callback, order): {!r}".format(hook))


def is_sub_path_of(path, parent):
	"""
	Tests if `path` is a sub path (or identical) to `path`.

	>>> is_sub_path_of("/a/b/c", "/a/b")
	True
	>>> is_sub_path_of("/a/b/c", "/a/b2")
	False
	>>> is_sub_path_of("/a/b/c", "/b/c")
	False
	>>> is_sub_path_of("/foo/bar/../../a/b/c", "/a/b")
	True
	>>> is_sub_path_of("/a/b", "/a/b")
	True
	"""
	rel_path = os.path.relpath(os.path.realpath(path),
	                           os.path.realpath(parent))
	return not (rel_path == os.pardir or
	            rel_path.startswith(os.pardir + os.sep))


def is_editable_install(install_dir, package, module, location):
	package_link = os.path.join(install_dir, "{}.egg-link".format(package))
	if os.path.isfile(package_link):
		expected_target = os.path.normcase(os.path.realpath(location))
		try:
			with io.open(package_link, 'rt', encoding='utf-8') as f:
				contents = f.readlines()
			for line in contents:
				target = os.path.normcase(os.path.realpath(os.path.join(line.strip(), module)))
				if target == expected_target:
					return True
		except Exception:
			raise # TODO really ignore this?
			pass
	return False


class EntryPointMetadata(pkginfo.Distribution):
	def __init__(self, entry_point):
		self.entry_point = entry_point
		self.extractMetadata()

	def read(self):
		import warnings

		metadata_files = ("METADATA",  # wheel
		                  "PKG-INFO")  # egg

		if self.entry_point and self.entry_point.dist:
			for metadata_file in metadata_files:
				try:
					return self.entry_point.dist.get_metadata(metadata_file)
				except (IOError, OSError):
					# file not found, metadata file might be missing, ignore
					# IOError: file not found in Py2
					# OSError (specifically FileNotFoundError): file not found in Py3
					pass

		warnings.warn('No package metadata found for package {}'.format(self.entry_point.module_name))


class Plugin(object):
	"""
	The parent class of all plugin implementations.

	.. attribute:: _identifier

	   The identifier of the plugin. Injected by the plugin core system upon initialization of the implementation.

	.. attribute:: _plugin_name

	   The name of the plugin. Injected by the plugin core system upon initialization of the implementation.

	.. attribute:: _plugin_version

	   The version of the plugin. Injected by the plugin core system upon initialization of the implementation.

	.. attribute:: _basefolder

	   The base folder of the plugin. Injected by the plugin core system upon initialization of the implementation.

	.. attribute:: _logger

	   The logger instance to use, with the logging name set to the :attr:`PluginManager.logging_prefix` of the
	   :class:`PluginManager` concatenated with :attr:`_identifier`. Injected by the plugin core system upon
	   initialization of the implementation.
	"""

	def __init__(self):
		self._identifier = None
		self._plugin_name = None
		self._plugin_version = None
		self._basefolder = None
		self._logger = None

	def initialize(self):
		"""
		Called by the plugin core after performing all injections. Override this to initialize your implementation.
		"""
		pass

	def on_plugin_enabled(self):
		pass

	def on_plugin_disabled(self):
		pass

class RestartNeedingPlugin(Plugin):
	"""
	Mixin for plugin types that need a restart after enabling/disabling them.
	"""

class SortablePlugin(Plugin):
	"""
	Mixin for plugin types that are sortable.
	"""

	def get_sorting_key(self, context=None):
		"""
		Returns the sorting key to use for the implementation in the specified ``context``.

		May return ``None`` if order is irrelevant.

		Implementations returning None will be ordered by plugin identifier
		after all implementations which did return a sorting key value that was
		not None sorted by that.

		Arguments:
		    context (str): The sorting context for which to provide the
		        sorting key value.

		Returns:
		    int or None: An integer signifying the sorting key value of the plugin
		        (sorting will be done ascending), or None if the implementation
		        doesn't care about calling order.
		"""
		return None

class PluginNeedsRestart(Exception):
	def __init__(self, name):
		Exception.__init__(self)
		self.name = name
		self.message = "Plugin {name} cannot be enabled or disabled after system startup".format(**locals())

class PluginLifecycleException(Exception):
	def __init__(self, name, reason, message):
		Exception.__init__(self)
		self.name = name
		self.reason = reason

		self.message = message.format(**locals())

	def __str__(self):
		return self.message

class PluginCantInitialize(PluginLifecycleException):
	def __init__(self, name, reason):
		PluginLifecycleException.__init__(self, name, reason, "Plugin {name} cannot be initialized: {reason}")

class PluginCantEnable(PluginLifecycleException):
	def __init__(self, name, reason):
		PluginLifecycleException.__init__(self, name, reason, "Plugin {name} cannot be enabled: {reason}")

class PluginCantDisable(PluginLifecycleException):
	def __init__(self, name, reason):
		PluginLifecycleException.__init__(self, name, reason, "Plugin {name} cannot be disabled: {reason}")



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import logging.handlers
import os
import re
import time

# noinspection PyCompatibility
import concurrent.futures


class AsyncLogHandlerMixin(logging.Handler):
	def __init__(self, *args, **kwargs):
		self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
		super(AsyncLogHandlerMixin, self).__init__(*args, **kwargs)

	def emit(self, record):
		try:
			self._executor.submit(self._emit, record)
		except Exception:
			self.handleError(record)

	def _emit(self, record):
		# noinspection PyUnresolvedReferences
		super(AsyncLogHandlerMixin, self).emit(record)


class CleaningTimedRotatingFileHandler(logging.handlers.TimedRotatingFileHandler):

	def __init__(self, *args, **kwargs):
		kwargs["encoding"] = kwargs.get("encoding", "utf-8")

		super(CleaningTimedRotatingFileHandler, self).__init__(*args, **kwargs)

		# clean up old files on handler start
		if self.backupCount > 0:
			for s in self.getFilesToDelete():
				os.remove(s)


class OctoPrintLogHandler(AsyncLogHandlerMixin, CleaningTimedRotatingFileHandler):
	rollover_callbacks = []

	def __init__(self, *args, **kwargs):
		kwargs["encoding"] = kwargs.get("encoding", "utf-8")
		super(OctoPrintLogHandler, self).__init__(*args, **kwargs)

	@classmethod
	def registerRolloverCallback(cls, callback, *args, **kwargs):
		cls.rollover_callbacks.append((callback, args, kwargs))

	def doRollover(self):
		super(OctoPrintLogHandler, self).doRollover()

		for rcb in self.rollover_callbacks:
			callback, args, kwargs = rcb
			callback(*args, **kwargs)


class OctoPrintStreamHandler(AsyncLogHandlerMixin, logging.StreamHandler):
	pass


class SerialLogHandler(AsyncLogHandlerMixin, logging.handlers.RotatingFileHandler):

	do_rollover = False
	suffix_template = "%Y-%m-%d_%H-%M-%S"
	file_pattern = re.compile(r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}$")

	@classmethod
	def on_open_connection(cls):
		cls.do_rollover = True

	def __init__(self, *args, **kwargs):
		kwargs["encoding"] = kwargs.get("encoding", "utf-8")
		super(SerialLogHandler, self).__init__(*args, **kwargs)
		self.cleanupFiles()

	def shouldRollover(self, record):
		return self.do_rollover

	def getFilesToDelete(self):
		"""
		Determine the files to delete when rolling over.
		"""
		dirName, baseName = os.path.split(self.baseFilename)
		fileNames = os.listdir(dirName)
		result = []
		prefix = baseName + "."
		plen = len(prefix)
		for fileName in fileNames:
			if fileName[:plen] == prefix:
				suffix = fileName[plen:]
				if type(self).file_pattern.match(suffix):
					result.append(os.path.join(dirName, fileName))
		result.sort()
		if len(result) < self.backupCount:
			result = []
		else:
			result = result[:len(result) - self.backupCount]
		return result

	def cleanupFiles(self):
		if self.backupCount > 0:
			for path in self.getFilesToDelete():
				os.remove(path)

	def doRollover(self):
		self.do_rollover = False

		if self.stream:
			self.stream.close()
			self.stream = None

		if os.path.exists(self.baseFilename):
			# figure out creation date/time to use for file suffix
			t = time.localtime(os.stat(self.baseFilename).st_mtime)
			dfn = self.baseFilename + "." + time.strftime(type(self).suffix_template, t)
			if os.path.exists(dfn):
				os.remove(dfn)
			os.rename(self.baseFilename, dfn)

		self.cleanupFiles()
		if not self.delay:
			self.stream = self._open()

class RecordingLogHandler(logging.Handler):
	def __init__(self, target=None, *args, **kwargs):
		super(RecordingLogHandler, self).__init__(*args, **kwargs)
		self._buffer = []
		self._target = target

	def emit(self, record):
		self._buffer.append(record)

	def setTarget(self, target):
		self._target = target

	def flush(self):
		if not self._target:
			return

		self.acquire()
		try:
			for record in self._buffer:
				self._target.handle(record)
			self._buffer = []
		finally:
			self.release()

	def close(self):
		self.flush()
		self.acquire()
		try:
			self._buffer = []
		finally:
			self.release()

	def __len__(self):
		return len(self._buffer)


# noinspection PyAbstractClass
class CombinedLogHandler(logging.Handler):
	def __init__(self, *handlers):
		logging.Handler.__init__(self)
		self._handlers = handlers

	def setHandlers(self, *handlers):
		self._handlers = handlers

	def handle(self, record):
		self.acquire()
		try:
			if self._handlers:
				for handler in self._handlers:
					handler.handle(record)
		finally:
			self.release()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import click
click.disable_unicode_literals_warning = True

#~~ "octoprint util" commands

@click.group()
def analysis_commands():
	pass

@analysis_commands.group(name="analysis")
def util():
	"""Analysis tools."""
	pass


@util.command(name="gcode")
@click.option("--throttle", "throttle", type=float, default=None)
@click.option("--throttle-lines", "throttle_lines", type=int, default=None)
@click.option("--speed-x", "speedx", type=float, default=6000)
@click.option("--speed-y", "speedy", type=float, default=6000)
@click.option("--speed-z", "speedz", type=float, default=300)
@click.option("--offset", "offset", type=(float, float), multiple=True)
@click.option("--max-t", "maxt", type=int, default=10)
@click.option("--g90-extruder", "g90_extruder", is_flag=True)
@click.option("--progress", "progress", is_flag=True)
@click.argument("path", type=click.Path())
def gcode_command(path, speedx, speedy, speedz, offset, maxt, throttle, throttle_lines, g90_extruder, progress):
	"""Runs a GCODE file analysis."""

	import time
	import yaml
	from octoprint.util import monotonic_time
	from octoprint.util.gcodeInterpreter import gcode

	throttle_callback = None
	if throttle:
		def throttle_callback(filePos, readBytes):
			if filePos % throttle_lines == 0:
				# only apply throttle every $throttle_lines lines
				time.sleep(throttle)

	offsets = offset
	if offsets is None:
		offsets = []
	elif isinstance(offset, tuple):
		offsets = list(offsets)
	offsets = [(0, 0)] + offsets
	if len(offsets) < maxt:
		offsets += [(0, 0)] * (maxt - len(offsets))

	start_time = monotonic_time()

	progress_callback = None
	if progress:
		def progress_callback(percentage):
			click.echo("PROGRESS:{}".format(percentage))
	interpreter = gcode(progress_callback=progress_callback)

	interpreter.load(path,
					 speedx=speedx,
					 speedy=speedy,
					 offsets=offsets,
					 throttle=throttle_callback,
					 max_extruders=maxt,
					 g90_extruder=g90_extruder)

	click.echo("DONE:{}s".format(monotonic_time() - start_time))
	click.echo("RESULTS:")
	click.echo(yaml.safe_dump(interpreter.get_result(), default_flow_style=False, indent=4, allow_unicode=True))

if __name__ == "__main__":
	gcode_command()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import flask
import io
import os
import sarge

from flask_babel import gettext
from octoprint.util import RepeatedTimer

from octoprint.access.permissions import Permissions
from octoprint.access.groups import USER_GROUP

import octoprint.plugin
import octoprint.events

_PROC_DT_MODEL_PATH = "/proc/device-tree/model"
_OCTOPI_VERSION_PATH = "/etc/octopi_version"
_VCGENCMD_THROTTLE = "/usr/bin/vcgencmd get_throttled"

_CHECK_INTERVAL_OK = 300
_CHECK_INTERVAL_THROTTLED = 30

__LOCAL_DEBUG = False

if __LOCAL_DEBUG:
	### mocks & settings for local debugging
	import sys
	base = os.path.realpath(os.path.join(os.path.dirname(__file__), "..", "..", "..", "..", "tests", "plugins", "pi_support", "fakes"))
	_PROC_DT_MODEL_PATH = os.path.join(base, "fake_model.txt")
	_OCTOPI_VERSION_PATH = os.path.join(base, "fake_octopi.txt")
	_VCGENCMD_THROTTLE = "{} {}".format(sys.executable, os.path.join(base, "fake_vcgencmd.py"))
	import itertools
	_VCGENCMD_OUTPUT = itertools.chain(iter(("0x0", "0x0", "0x50005", "0x50000", "0x70007")), itertools.repeat("0x70005"))

	_CHECK_INTERVAL_OK = 10
	_CHECK_INTERVAL_THROTTLED = 5


# see https://www.raspberrypi.org/forums/viewtopic.php?f=63&t=147781&start=50#p972790
_FLAG_UNDERVOLTAGE = 1 << 0
_FLAG_FREQ_CAPPED = 1 << 1
_FLAG_THROTTLED = 1 << 2
_FLAG_PAST_UNDERVOLTAGE = 1 << 16
_FLAG_PAST_FREQ_CAPPED = 1 << 17
_FLAG_PAST_THROTTLED = 1 << 18

class ThrottleState(object):
	@classmethod
	def from_value(cls, value=0):
		if value == 0:
			return ThrottleState()

		kwargs = dict(undervoltage=_FLAG_UNDERVOLTAGE & value == _FLAG_UNDERVOLTAGE,
		              freq_capped=_FLAG_FREQ_CAPPED & value == _FLAG_FREQ_CAPPED,
		              throttled=_FLAG_THROTTLED & value == _FLAG_THROTTLED,
		              past_undervoltage=_FLAG_PAST_UNDERVOLTAGE & value == _FLAG_PAST_UNDERVOLTAGE,
		              past_freq_capped=_FLAG_PAST_FREQ_CAPPED & value == _FLAG_PAST_FREQ_CAPPED,
		              past_throttled=_FLAG_PAST_THROTTLED & value == _FLAG_PAST_THROTTLED,
		              raw_value=value)
		return ThrottleState(**kwargs)

	def __init__(self, **kwargs):
		self._raw_value = kwargs.get('raw_value', -1)
		self._undervoltage = False
		self._freq_capped = False
		self._throttled = False
		self._past_undervoltage = False
		self._past_freq_capped = False
		self._past_throttled = False

		for key, value in kwargs.items():
			local_key = "_{}".format(key)
			if hasattr(self, local_key) and isinstance(value, bool):
				setattr(self, local_key, value)

	@property
	def undervoltage(self):
		return self.current_undervoltage or self.past_undervoltage

	@property
	def current_undervoltage(self):
		return self._undervoltage

	@property
	def past_undervoltage(self):
		return self._past_undervoltage

	@property
	def overheat(self):
		return self.current_overheat or self.past_overheat

	@property
	def current_overheat(self):
		return self._freq_capped

	@property
	def past_overheat(self):
		return self._past_freq_capped

	@property
	def issue(self):
		return self.current_issue or self.past_issue

	@property
	def current_issue(self):
		return self._undervoltage or self._freq_capped or self._throttled

	@property
	def past_issue(self):
		return self._past_undervoltage or self._past_freq_capped or self._past_throttled

	def __eq__(self, other):
		if not isinstance(other, ThrottleState):
			return False

		return self._undervoltage == other._undervoltage \
		       and self._freq_capped == other._freq_capped \
		       and self._throttled == other._throttled \
		       and self._past_undervoltage == other._past_undervoltage \
		       and self._past_freq_capped == other._past_freq_capped \
		       and self._past_throttled == other._past_throttled

	def as_dict(self):
		return dict(raw_value=self._raw_value,
		            current_undervoltage=self.current_undervoltage,
		            past_undervoltage=self.past_undervoltage,
		            current_overheat=self.current_overheat,
		            past_overheat=self.past_overheat,
		            current_issue=self.current_issue,
		            past_issue=self.past_issue)


_proc_dt_model = None
def get_proc_dt_model():
	global _proc_dt_model

	if _proc_dt_model is None:
		with io.open(_PROC_DT_MODEL_PATH, 'rt', encoding='utf-8') as f:
			_proc_dt_model = f.readline().strip(" \t\r\n\0")

	return _proc_dt_model


def get_vcgencmd_throttled_state(command):
	if __LOCAL_DEBUG:
		output = "throttled={}".format(next(_VCGENCMD_OUTPUT))  # mock for local debugging
	else:
		output = sarge.get_stdout(command)

	if not "throttled=0x" in output:
		raise ValueError("cannot parse \"{}\" output: {}".format(command, output))

	value = output[len("throttled="):].strip(" \t\r\n\0")
	value = int(value, 0)
	return ThrottleState.from_value(value)


def is_octopi():
	return os.path.exists(_OCTOPI_VERSION_PATH)


_octopi_version = None
def get_octopi_version():
	global _octopi_version

	if _octopi_version is None:
		with io.open(_OCTOPI_VERSION_PATH, 'rt', encoding='utf-8') as f:
			_octopi_version = f.readline().strip(" \t\r\n\0")

	return _octopi_version


class PiSupportPlugin(octoprint.plugin.EnvironmentDetectionPlugin,
                      octoprint.plugin.SimpleApiPlugin,
                      octoprint.plugin.AssetPlugin,
                      octoprint.plugin.TemplatePlugin,
                      octoprint.plugin.StartupPlugin,
                      octoprint.plugin.SettingsPlugin):

	# noinspection PyMissingConstructor
	def __init__(self):
		self._throttle_state = ThrottleState()
		self._throttle_check = None
		self._throttle_undervoltage = False
		self._throttle_overheat = False
		self._throttle_functional = True

	# Additional permissions hook

	def get_additional_permissions(self):
		return [
			dict(key="STATUS",
			     name="Status",
			     description=gettext("Allows to check for the Pi's throttling status and environment info"),
			     roles=["check"],
			     default_groups=[USER_GROUP])
		]

	#~~ EnvironmentDetectionPlugin

	def get_additional_environment(self):
		result = dict(model=get_proc_dt_model())

		if is_octopi():
			result.update(dict(octopi_version=get_octopi_version()))

		return result

	#~~ SimpleApiPlugin

	def on_api_get(self, request):
		if not Permissions.PLUGIN_PI_SUPPORT_STATUS.can():
			return flask.abort(403)
		result = dict(throttle_state=self._throttle_state.as_dict())
		result.update(self.get_additional_environment())
		return flask.jsonify(**result)

	#~~ AssetPlugin

	def get_assets(self):
		return dict(
			js=["js/pi_support.js"],
			clientjs=["clientjs/pi_support.js"],
			css=["css/pi_support.css"]
		)

	#~~ TemplatePlugin

	def get_template_configs(self):
		configs = [dict(type="settings", name=gettext("Pi Support"), template="pi_support_settings.jinja2", custom_bindings=False)]

		if is_octopi():
			configs.append(dict(type="about", name="About OctoPi", template="pi_support_about_octopi.jinja2"))

		return configs

	def get_template_vars(self):
		return self.get_additional_environment()

	#~~ StartupPlugin

	def on_startup(self, *args, **kwargs):
		if self._settings.get_boolean(["vcgencmd_throttle_check_enabled"]):
			self._check_throttled_state()
			self._throttle_check = RepeatedTimer(self._check_throttled_state_interval,
			                                     self._check_throttled_state,
			                                     condition=self._check_throttled_state_condition)
			self._throttle_check.start()

	#~~ SettingsPlugin

	def get_settings_defaults(self):
		return dict(vcgencmd_throttle_check_enabled=True,
		            vcgencmd_throttle_check_command=_VCGENCMD_THROTTLE)

	def get_settings_restricted_paths(self):
		return dict(admin=[["vcgencmd_throttle_check_enabled"], ["vcgencmd_throttle_check_command"]])

	#~~ Helpers

	def _check_throttled_state_interval(self):
		if self._throttle_state.current_issue:
			return _CHECK_INTERVAL_THROTTLED
		else:
			return _CHECK_INTERVAL_OK

	def _check_throttled_state_condition(self):
		return self._throttle_functional

	def get_throttle_state(self, run_now=False):
		"""Exposed as public helper."""
		if run_now:
			self._check_throttled_state()

		if not self._throttle_functional:
			return False

		return self._throttle_state.as_dict()

	def _check_throttled_state(self):
		command = self._settings.get(["vcgencmd_throttle_check_command"])

		self._logger.debug("Retrieving throttle state via \"{}\"".format(command))
		try:
			state = get_vcgencmd_throttled_state(command)
		except ValueError:
			self._logger.warning("Fetching the current throttle state via \"{}\" doesn't work".format(command))
			self._throttle_functional = False
			return

		if self._throttle_state == state:
			# no change
			return

		self._throttle_state = state

		if (not self._throttle_undervoltage and self._throttle_state.undervoltage) \
				or (not self._throttle_overheat and self._throttle_state.overheat):
			message = "This Raspberry Pi is reporting problems that might lead to bad performance or errors caused " \
			          "by overheating or insufficient power."

			if self._throttle_state.undervoltage:
				self._throttle_undervoltage = True
				message += "\n!!! UNDERVOLTAGE REPORTED !!! Make sure that the power supply and power cable are " \
				           "capable of supplying enough voltage and current to your Pi."

			if self._throttle_state.overheat:
				self._throttle_overheat = True
				message += "\n!!! FREQUENCY CAPPING DUE TO OVERHEATING REPORTED !!! Improve cooling on the Pi's " \
				           "CPU and GPU."

			self._logger.warning(message)

		self._plugin_manager.send_plugin_message(self._identifier, dict(type="throttle_state",
		                                                                state=self._throttle_state.as_dict()))

		# noinspection PyUnresolvedReferences
		self._event_bus.fire(octoprint.events.Events.PLUGIN_PI_SUPPORT_THROTTLE_STATE, self._throttle_state.as_dict())


def register_custom_events(*args, **kwargs):
	return ["throttle_state",]


__plugin_name__ = "Pi Support Plugin"
__plugin_author__ = "Gina Häußge"
__plugin_description__ = "Provides additional information about your Pi in the UI."
__plugin_disabling_discouraged__ = gettext("Without this plugin OctoPrint will no longer be able to "
                                           "provide additional information about your Pi, "
                                           "which will make it more tricky to help you if you need support.")
__plugin_license__ = "AGPLv3"

def __plugin_check__():
	try:
		proc_dt_model = get_proc_dt_model()
		if proc_dt_model is None:
			return False
	except Exception:
		return False

	return "raspberry pi" in proc_dt_model.lower()

def __plugin_load__():
	plugin = PiSupportPlugin()
	global __plugin_implementation__
	__plugin_implementation__ = plugin

	global __plugin_hooks__
	__plugin_hooks__ = {
		"octoprint.events.register_custom_events": register_custom_events,
		"octoprint.access.permissions": __plugin_implementation__.get_additional_permissions
	}

	global __plugin_helpers__
	__plugin_helpers__ = dict(get_throttled=plugin.get_throttle_state)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


def commands(cli_group, pass_octoprint_ctx, *args, **kwargs):
	import click
	click.disable_unicode_literals_warning = True
	import sys
	import requests.exceptions
	from octoprint.cli.client import create_client, client_options

	@click.command("check")
	@click.option("--force", is_flag=True, help="Ignore the cache for the update check")
	@click.option("--only-new", is_flag=True, help="Only show entries with updates available")
	@client_options
	@click.argument("targets", nargs=-1)
	def check_command(force, only_new, apikey, host, port, httpuser, httppass, https, prefix, targets):
		"""
		Check for updates.

		If any TARGETs are provided, only those components will be checked.

		\b
		Examples:
		- octoprint plugins softwareupdate:check
		    This will check all components for available updates,
		    utilizing cached version information.
		- octoprint plugins softwareupdate:check --force
		    This will check all components for available updates,
		    ignoring any cached version information even if it's
		    still valid.
		- octoprint plugins softwareupdate:check octoprint
		    This will only check OctoPrint itself for available
		    updates.
		"""
		params = dict(force=force)
		if targets:
			params["check"] = ",".join(targets)

		client = create_client(settings=cli_group.settings,
		                       apikey=apikey,
		                       host=host,
		                       port=port,
		                       httpuser=httpuser,
		                       httppass=httppass,
		                       https=https,
		                       prefix=prefix)

		r = client.get("plugin/softwareupdate/check", params=params)
		try:
			r.raise_for_status()
		except requests.exceptions.HTTPError as e:
			click.echo("Could not get update information from server, got {}".format(e))
			sys.exit(1)

		data = r.json()
		status = data["status"]
		information = data["information"]

		lines = []
		octoprint_line = None
		for key, info in information.items():
			status_text = "Up to date"
			if info["updateAvailable"]:
				if info["updatePossible"]:
					status_text = "Update available"
				else:
					status_text = "Update available (manual)"

			elif only_new:
				continue

			line = "{} (target: {})\n\tInstalled: {}\n\tAvailable: {}\n\t=> {}".format(info["displayName"],
			                                                                           key,
			                                                                           info["information"]["local"]["name"],
			                                                                           info["information"]["remote"]["name"],
			                                                                           status_text)
			if key == "octoprint":
				octoprint_line = line
			else:
				lines.append(line)

		lines.sort()
		if octoprint_line:
			lines = [octoprint_line] + lines

		for line in lines:
			click.echo(line)

		click.echo()
		if status == "current":
			click.echo("Everything is up to date")
		else:
			click.echo("There are updates available!")


	@click.command("update")
	@click.option("--force", is_flag=True, help="Update even if already up to date")
	@client_options
	@click.argument("targets", nargs=-1)
	def update_command(force, apikey, host, port, httpuser, httppass, https, prefix, targets):
		"""
		Apply updates.

		If any TARGETs are provided, only those components will be updated.

		\b
		Examples:
		- octoprint plugins softwareupdate:update
		    This will update all components with a pending update
		    that can be updated.
		- octoprint plugins softwareupdate:update --force
		    This will force an update of all registered components
		    that can be updated, even if they don't have an updated
		    pending.
		- octoprint plugins softwareupdate:update octoprint
		    This will only update OctoPrint and leave any further
		    components with pending updates at their current versions.
		"""

		data = dict(force=force)
		if targets:
			data["check"] = targets

		client = create_client(settings=cli_group.settings,
		                       apikey=apikey,
		                       host=host,
		                       port=port,
		                       httpuser=httpuser,
		                       httppass=httppass,
		                       https=https,
		                       prefix=prefix)

		flags = dict(
			waiting_for_restart=False,
			seen_close=False
		)

		def on_message(ws, msg_type, msg):
			if msg_type != "plugin" or msg["plugin"] != "softwareupdate":
				return

			plugin_message = msg["data"]
			if not "type" in plugin_message:
				return

			plugin_message_type = plugin_message["type"]
			plugin_message_data = plugin_message["data"]

			if plugin_message_type == "updating":
				click.echo("Updating {} to {}...".format(plugin_message_data.get("name", "unknown"), plugin_message_data.get("version", "n/a")))

			elif plugin_message_type == "update_failed":
				click.echo("\t... failed :(")

			elif plugin_message_type == "loglines" and "loglines" in plugin_message_data:
				for entry in plugin_message_data["loglines"]:
					prefix = ">>> " if entry["stream"] == "call" else ""
					error = entry["stream"] == "stderr"
					click.echo("\t{}{}".format(prefix, entry["line"].replace("\n", "\n\t")), err=error)

			elif plugin_message_type == "success" or plugin_message_type == "restart_manually":
				results = plugin_message_data["results"] if "results" in plugin_message_data else dict()
				if results:
					click.echo("The update finished successfully.")
					if plugin_message_type == "restart_manually":
						click.echo("Please restart the OctoPrint server.")
				else:
					click.echo("No update necessary")
				ws.close()

			elif plugin_message_type == "restarting":
				flags["waiting_for_restart"] = True
				click.echo("Restarting to apply changes...")

			elif plugin_message_type == "failure":
				click.echo("Error")
				ws.close()

		def on_open(ws):
			if flags["waiting_for_restart"] and flags["seen_close"]:
				click.echo(" Reconnected!")
			else:
				click.echo("Connected to server...")

		def on_close(ws):
			if flags["waiting_for_restart"] and flags["seen_close"]:
				click.echo(".", nl=False)
			else:
				flags["seen_close"] = True
				click.echo("Disconnected from server...")

		socket = client.create_socket(on_message=on_message,
		                              on_open=on_open,
		                              on_close=on_close)

		r = client.post_json("plugin/softwareupdate/update", data=data)
		try:
			r.raise_for_status()
		except requests.exceptions.HTTPError as e:
			click.echo("Could not get update information from server, got {}".format(e))
			sys.exit(1)

		data = r.json()
		to_be_updated = data["order"]
		checks = data["checks"]
		click.echo("Update in progress, updating:")
		for name in to_be_updated:
			click.echo("\t{}".format(name if not name in checks else checks[name]))

		socket.wait()

		if flags["waiting_for_restart"]:
			if socket.reconnect(timeout=60):
				click.echo("The update finished successfully.")
			else:
				click.echo("The update finished successfully but the server apparently didn't restart as expected.")
				click.echo("Please restart the OctoPrint server.")

	return [check_command, update_command]


